"""
"""

# Created on 2015.02.3
#
# Author: Giovanni Cannata
#
# Copyright 2015, 2016 Giovanni Cannata
#
# This file is part of ldap3.
#
# ldap3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ldap3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ldap3 in the COPYING and COPYING.LESSER files.
# If not, see <http://www.gnu.org/licenses/>.


import unittest

from ldap3 import Server, Connection, MOCK_SYNC, MODIFY_ADD, MODIFY_REPLACE, MODIFY_DELETE, OFFLINE_EDIR_8_8_8,\
    BASE, LEVEL, SUBTREE, AUTO_BIND_NO_TLS, NONE
from ldap3.core.exceptions import LDAPInvalidCredentialsResult, LDAPNoSuchObjectResult
from ldap3.protocol.rfc4512 import SchemaInfo, DsaInfo
from ldap3.protocol.schemas.edir888 import edir_8_8_8_dsa_info, edir_8_8_8_schema
from test import random_id

testcase_id = random_id()


class Test(unittest.TestCase):
    def setUp(self):
        # The mock server can be defined in two different ways, so tests are duplicated, connection_3 is without schema
        schema = SchemaInfo.from_json(edir_8_8_8_schema)
        info = DsaInfo.from_json(edir_8_8_8_dsa_info, schema)
        server_1 = Server.from_definition('MockSyncServer', info, schema)
        self.connection_1 = Connection(server_1, user='cn=user1,ou=test,o=lab', password='test1111', client_strategy=MOCK_SYNC)
        self.connection_1b = Connection(server_1, user='cn=user1,ou=test,o=lab', password='test1111', client_strategy=MOCK_SYNC)
        self.connection_1c = Connection(server_1, user='cn=user1,ou=test,o=lab', password='test1111', client_strategy=MOCK_SYNC, raise_exceptions=True)
        server_2 = Server('dummy', get_info=OFFLINE_EDIR_8_8_8)
        self.connection_2 = Connection(server_2, user='cn=user2,ou=test,o=lab', password='test2222', client_strategy=MOCK_SYNC)
        self.connection_2b = Connection(server_2, user='cn=user2,ou=test,o=lab', password='test2222', client_strategy=MOCK_SYNC)
        self.connection_2c = Connection(server_2, user='cn=user2,ou=test,o=lab', password='test2222', client_strategy=MOCK_SYNC, raise_exceptions=True)
        server_3 = Server('dummy')  # no schema
        self.connection_3 = Connection(server_3, user='cn=user3,ou=test,o=lab', password='test3333', client_strategy=MOCK_SYNC)
        self.connection_3b = Connection(server_3, user='cn=user3,ou=test,o=lab', password='test3333', client_strategy=MOCK_SYNC)
        self.connection_3c = Connection(server_3, user='cn=user3,ou=test,o=lab', password='test3333', client_strategy=MOCK_SYNC, raise_exceptions=True)
        # creates fixtures
        self.connection_1.strategy.add_entry('cn=user0,o=lab', {'userPassword': 'test0000', 'sn': 'user0_sn', 'revision': 0})
        self.connection_2.strategy.add_entry('cn=user0,o=lab', {'userPassword': 'test0000', 'sn': 'user0_sn', 'revision': 0})
        self.connection_3.strategy.add_entry('cn=user0,o=lab', {'userPassword': 'test0000', 'sn': 'user0_sn', 'revision': 0})
        self.connection_1.strategy.add_entry('cn=user1,ou=test,o=lab', {'userPassword': 'test1111', 'sn': 'user1_sn', 'revision': 1})
        self.connection_2.strategy.add_entry('cn=user1,ou=test,o=lab', {'userPassword': 'test1111', 'sn': 'user1_sn', 'revision': 1})
        self.connection_3.strategy.add_entry('cn=user1,ou=test,o=lab', {'userPassword': 'test1111', 'sn': 'user1_sn', 'revision': 1})
        self.connection_1.strategy.add_entry('cn=user2,ou=test,o=lab', {'userPassword': 'test2222', 'sn': 'user2_sn', 'revision': 2})
        self.connection_2.strategy.add_entry('cn=user2,ou=test,o=lab', {'userPassword': 'test2222', 'sn': 'user2_sn', 'revision': 2})
        self.connection_3.strategy.add_entry('cn=user2,ou=test,o=lab', {'userPassword': 'test2222', 'sn': 'user2_sn', 'revision': 2})
        self.connection_1.strategy.add_entry('cn=user3,ou=test,o=lab', {'userPassword': 'test3333', 'sn': 'user3_sn', 'revision': 3})
        self.connection_2.strategy.add_entry('cn=user3,ou=test,o=lab', {'userPassword': 'test3333', 'sn': 'user3_sn', 'revision': 3})
        self.connection_3.strategy.add_entry('cn=user3,ou=test,o=lab', {'userPassword': 'test3333', 'sn': 'user3_sn', 'revision': 3})
        self.connection_1.strategy.add_entry('cn=user4,ou=test,o=lab', {'userPassword': 'test4444', 'sn': 'user4_sn', 'revision': 4, 'title': ['title1', 'title2', 'title3']})
        self.connection_2.strategy.add_entry('cn=user4,ou=test,o=lab', {'userPassword': 'test4444', 'sn': 'user4_sn', 'revision': 4, 'title': ['title1', 'title2', 'title3']})
        self.connection_3.strategy.add_entry('cn=user4,ou=test,o=lab', {'userPassword': 'test4444', 'sn': 'user4_sn', 'revision': 4, 'title': ['title1', 'title2', 'title3']})

    def tearDown(self):
        self.connection_1.unbind()
        self.assertFalse(self.connection_1.bound)
        self.connection_2.unbind()
        self.assertFalse(self.connection_2.bound)
        self.connection_3.unbind()
        self.assertFalse(self.connection_3.bound)

    def test_open_1(self):
        self.connection_1.open()
        self.assertFalse(self.connection_1.closed)

    def test_open_2(self):
        self.connection_2.open()
        self.assertFalse(self.connection_2.closed)

    def test_open_3(self):
        self.connection_3.open()
        self.assertFalse(self.connection_3.closed)

    def test_bind_1(self):
        self.connection_1.open()
        self.connection_1.bind()
        self.assertTrue(self.connection_1.bound)

    def test_bind_2(self):
        self.connection_2.open()
        self.connection_2.bind()
        self.assertTrue(self.connection_2.bound)

    def test_bind_3(self):
        self.connection_3.open()
        self.connection_3.bind()
        self.assertTrue(self.connection_3.bound)

    def test_invalid_bind_1(self):
        self.connection_1.password = 'wrong'
        self.connection_1.open()
        self.connection_1.bind()
        self.assertFalse(self.connection_1.bound)

    def test_invalid_bind_2(self):
        self.connection_2.password = 'wrong'
        self.connection_2.open()
        self.connection_2.bind()
        self.assertFalse(self.connection_2.bound)

    def test_invalid_bind_3(self):
        self.connection_3.password = 'wrong'
        self.connection_3.open()
        self.connection_3.bind()
        self.assertFalse(self.connection_3.bound)

    def test_invalid_bind_exception_1(self):
        self.connection_1c.password = 'wrong'
        self.connection_1c.open()
        try:
            self.connection_1c.bind()
            self.fail('exception not raised')
        except LDAPInvalidCredentialsResult:
            pass

    def test_invalid_bind_exception_2(self):
        self.connection_2c.password = 'wrong'
        self.connection_2c.open()
        try:
            self.connection_2c.bind()
            self.fail('exception not raised')
        except LDAPInvalidCredentialsResult:
            pass

    def test_invalid_bind_exception_3(self):
        self.connection_3c.password = 'wrong'
        self.connection_3c.open()
        try:
            self.connection_3c.bind()
            self.fail('exception not raised')
        except LDAPInvalidCredentialsResult:
            pass

    def test_unbind_1(self):
        self.connection_1.open()
        self.assertFalse(self.connection_1.closed)
        self.connection_1.bind()
        self.assertTrue(self.connection_1.bound)
        self.connection_1.unbind()
        self.assertFalse(self.connection_1.bound)
        self.assertTrue(self.connection_1.closed)

    def test_unbind_2(self):
        self.connection_2.open()
        self.assertFalse(self.connection_2.closed)
        self.connection_2.bind()
        self.assertTrue(self.connection_2.bound)
        self.connection_2.unbind()
        self.assertFalse(self.connection_2.bound)
        self.assertTrue(self.connection_2.closed)

    def test_unbind_3(self):
        self.connection_3.open()
        self.assertFalse(self.connection_3.closed)
        self.connection_3.bind()
        self.assertTrue(self.connection_3.bound)
        self.connection_3.unbind()
        self.assertFalse(self.connection_3.bound)
        self.assertTrue(self.connection_3.closed)

    def test_add_user_1(self):
        self.connection_1.strategy.add_entry('cn=user4,ou=test,o=lab', {'userPassword': 'test4444'})
        self.connection_1.rebind('cn=user4,ou=test,o=lab', 'test4444')
        self.assertTrue(self.connection_1.bound)

    def test_add_user_2(self):
        self.connection_2.strategy.add_entry('cn=user4,ou=test,o=lab', {'userPassword': 'test4444'})
        self.connection_2.rebind('cn=user4,ou=test,o=lab', 'test4444')
        self.assertTrue(self.connection_2.bound)

    def test_add_user_3(self):
        self.connection_3.strategy.add_entry('cn=user4,ou=test,o=lab', {'userPassword': 'test4444'})
        self.connection_3.rebind('cn=user4,ou=test,o=lab', 'test4444')
        self.assertTrue(self.connection_3.bound)

    def test_remove_user_1(self):
        self.connection_1.strategy.remove_entry('cn=user2,ou=test,o=lab')
        self.connection_1.bind()
        self.assertTrue(self.connection_1.bound)
        self.connection_1.rebind('cn=user2,ou=test,o=lab', 'test9876')
        self.assertFalse(self.connection_1.bound)

    def test_remove_user_2(self):
        self.connection_2.strategy.remove_entry('cn=user1,ou=test,o=lab')
        self.connection_2.bind()
        self.assertTrue(self.connection_2.bound)
        self.connection_2.rebind('cn=user1,ou=test,o=lab', 'test9876')
        self.assertFalse(self.connection_2.bound)

    def test_remove_user_3(self):
        self.connection_3.strategy.remove_entry('cn=user1,ou=test,o=lab')
        self.connection_3.bind()
        self.assertTrue(self.connection_3.bound)
        self.connection_3.rebind('cn=user1,ou=test,o=lab', 'test1111')
        self.assertFalse(self.connection_3.bound)

    def test_delete_1(self):
        self.connection_1.bind()
        result = self.connection_1.delete('cn=user1,ou=test,o=lab')
        if not self.connection_1.strategy.sync:
            _, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result

        self.assertTrue(result['description'], 'success')
        self.assertTrue('cn=user1,ou=test,o=lab' not in self.connection_1.strategy.entries)

    def test_delete_2(self):
        self.connection_2.bind()
        result = self.connection_2.delete('cn=user2,ou=test,o=lab')
        if not self.connection_2.strategy.sync:
            _, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertTrue(['description'], 'success')
        self.assertTrue('cn=user2,ou=test,o=lab' not in self.connection_2.strategy.entries)

    def test_delete_3(self):
        self.connection_3.bind()
        result = self.connection_3.delete('cn=user3,ou=test,o=lab')
        if not self.connection_3.strategy.sync:
            _, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertTrue(result['description'], 'success')
        self.assertTrue('cn=user3,ou=test,o=lab' not in self.connection_3.strategy.entries)

    def test_add_entry_1(self):
        dn = 'cn=user5,ou=test,o=lab'
        self.connection_1.bind()
        result = self.connection_1.add(dn, ['inetOrgPerson', 'top'], {'sn': 'user5_sn'})
        if not self.connection_1.strategy.sync:
            _, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertTrue(result['description'], 'success')
        self.assertTrue(dn in self.connection_1.strategy.entries)
        self.connection_1b.bind()
        self.assertTrue(dn in self.connection_1b.strategy.entries)

    def test_add_entry_2(self):
        dn = 'cn=user5,ou=test,o=lab'
        self.connection_2.bind()
        result = self.connection_2.add(dn, ['inetOrgPerson', 'top'], {'sn': 'user5_sn'})
        if not self.connection_2.strategy.sync:
            _, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertTrue(result['description'], 'success')
        self.assertTrue(dn in self.connection_2.strategy.entries)
        self.connection_2b.bind()
        self.assertTrue(dn in self.connection_2b.strategy.entries)

    def test_add_entry_3(self):
        dn = 'cn=user5,ou=test,o=lab'
        self.connection_3.bind()
        result = self.connection_3.add(dn, ['inetOrgPerson', 'top'], {'sn': 'user5_sn'})
        if not self.connection_3.strategy.sync:
            _, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertTrue(result['description'], 'success')
        self.assertTrue(dn in self.connection_3.strategy.entries)
        self.connection_3b.bind()
        self.assertTrue(dn in self.connection_3b.strategy.entries)

    def test_add_entry_already_exists_1(self):
        dn = 'cn=user5,ou=test,o=lab'
        self.connection_1.bind()
        self.connection_1.strategy.add_entry(dn, {'objectClass': ['inetOrgPerson', 'top'], 'sn': 'user5_sn'})
        result = self.connection_1.add(dn, ['inetOrgPerson', 'top'], {'sn': 'user5_sn'})
        if not self.connection_1.strategy.sync:
            _, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertEqual(result['description'], 'entryAlreadyExists')

    def test_add_entry_already_exists_2(self):
        dn = 'cn=user5,ou=test,o=lab'
        self.connection_2.bind()
        self.connection_2.strategy.add_entry(dn, {'objectClass': ['inetOrgPerson', 'top'], 'sn': 'user5_sn'})
        result = self.connection_2.add(dn, ['inetOrgPerson', 'top'], {'sn': 'user5_sn'})
        if not self.connection_2.strategy.sync:
            _, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertEqual(result['description'], 'entryAlreadyExists')

    def test_add_entry_already_exists_3(self):
        dn = 'cn=user5,ou=test,o=lab'
        self.connection_3.bind()
        self.connection_3.strategy.add_entry(dn, {'objectClass': ['inetOrgPerson', 'top'], 'sn': 'user5_sn'})
        result = self.connection_3.add(dn, ['inetOrgPerson', 'top'], {'sn': 'user5_sn'})
        if not self.connection_3.strategy.sync:
            _, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertEqual(result['description'], 'entryAlreadyExists')

    def test_delete_entry_nonexisting_1(self):
        dn = 'cn=user5,ou=test,o=lab'
        self.connection_1.bind()
        self.assertTrue(dn not in self.connection_1.strategy.entries)
        result = self.connection_1.delete(dn)
        if not self.connection_1.strategy.sync:
            _, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertEqual(result['description'], 'noSuchObject')

    def test_delete_entry_nonexisting_2(self):
        dn = 'cn=user5,ou=test,o=lab'
        self.connection_2.bind()
        self.assertTrue(dn not in self.connection_2.strategy.entries)
        result = self.connection_2.delete(dn)
        if not self.connection_2.strategy.sync:
            _, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertEqual(result['description'], 'noSuchObject')

    def test_delete_entry_nonexisting_3(self):
        dn = 'cn=user5,ou=test,o=lab'
        self.connection_3.bind()
        self.assertTrue(dn not in self.connection_3.strategy.entries)
        result = self.connection_3.delete(dn)
        if not self.connection_3.strategy.sync:
            _, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertEqual(result['description'], 'noSuchObject')

    def test_compare_entry_1(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_1.bind()
        result = self.connection_1.compare(dn, 'sn', 'user4_sn')
        if not self.connection_1.strategy.sync:
            _, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertTrue(result['description'], 'compareTrue')
        result = self.connection_1.compare(dn, 'sn', 'bad_sn')
        if not self.connection_1.strategy.sync:
            _, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertTrue(result['description'], 'compareFalse')

    def test_compare_entry_2(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_2.bind()
        result = self.connection_2.compare(dn, 'sn', 'user4_sn')
        if not self.connection_2.strategy.sync:
            _, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertTrue(result['description'], 'compareTrue')
        result = self.connection_2.compare(dn, 'sn', 'bad_sn')
        if not self.connection_2.strategy.sync:
            _, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertTrue(result['description'], 'compareFalse')

    def test_compare_entry_3(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_3.bind()
        result = self.connection_3.compare(dn, 'sn', 'user4_sn')
        if not self.connection_3.strategy.sync:
            _, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertTrue(result['description'], 'compareTrue')
        result = self.connection_3.compare(dn, 'sn', 'bad_sn')
        if not self.connection_3.strategy.sync:
            _, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertTrue(result['description'], 'compareFalse')

    def test_move_dn_1(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_1.bind()
        result = self.connection_1.modify_dn(dn, relative_dn='cn=user4', new_superior='ou=moved,o=lab')
        if not self.connection_1.strategy.sync:
            _, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertTrue(result['description'], 'success')
        self.assertTrue('cn=user4,ou=moved,o=lab' in self.connection_1.strategy.entries)
        self.assertFalse(dn in self.connection_1.strategy.entries)

    def test_move_dn_2(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_2.bind()
        result = self.connection_2.modify_dn(dn, relative_dn='cn=user4', new_superior='ou=moved,o=lab')
        if not self.connection_2.strategy.sync:
            _, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertTrue(result['description'], 'success')
        self.assertTrue('cn=user4,ou=moved,o=lab' in self.connection_2.strategy.entries)
        self.assertFalse(dn in self.connection_2.strategy.entries)

    def test_move_dn_3(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_3.bind()
        result = self.connection_3.modify_dn(dn, relative_dn='cn=user4', new_superior='ou=moved,o=lab')
        if not self.connection_3.strategy.sync:
            _, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertTrue(result['description'], 'success')
        self.assertTrue('cn=user4,ou=moved,o=lab' in self.connection_3.strategy.entries)
        self.assertFalse(dn in self.connection_3.strategy.entries)

    def test_rename_dn_1(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_1.bind()
        result = self.connection_1.modify_dn(dn, relative_dn='cn=user_renamed')
        if not self.connection_1.strategy.sync:
            _, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertTrue(result['description'], 'success')
        self.assertTrue('cn=user_renamed,ou=test,o=lab' in self.connection_1.strategy.entries)
        self.assertFalse(dn in self.connection_1.strategy.entries)

    def test_rename_dn_2(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_2.bind()
        result = self.connection_2.modify_dn(dn, relative_dn='cn=user_renamed')
        if not self.connection_2.strategy.sync:
            _, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertTrue(result['description'], 'success')
        self.assertTrue('cn=user_renamed,ou=test,o=lab' in self.connection_2.strategy.entries)
        self.assertFalse(dn in self.connection_2.strategy.entries)

    def test_rename_dn_3(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_3.bind()
        result = self.connection_3.modify_dn(dn, relative_dn='cn=user_renamed')
        if not self.connection_3.strategy.sync:
            _, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertTrue(result['description'], 'success')
        self.assertTrue('cn=user_renamed,ou=test,o=lab' in self.connection_3.strategy.entries)
        self.assertFalse(dn in self.connection_3.strategy.entries)

    def test_modify_add_existing_singlevalue_1(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_1.bind()
        result = self.connection_1.modify(dn, {'sn': (MODIFY_ADD, ['sn_added'])})
        if not self.connection_1.strategy.sync:
            _, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertTrue(result['description'], 'success')
        self.assertTrue(b'sn_added' in self.connection_1.strategy.entries[dn]['sn'])

    def test_modify_add_existing_singlevalue_2(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_2.bind()
        result = self.connection_2.modify(dn, {'sn': (MODIFY_ADD, ['sn_added'])})
        if not self.connection_2.strategy.sync:
            _, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertTrue(result['description'], 'success')
        self.assertTrue(b'sn_added' in self.connection_2.strategy.entries[dn]['sn'])

    def test_modify_add_existing_singlevalue_3(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_3.bind()
        result = self.connection_3.modify(dn, {'sn': (MODIFY_ADD, ['sn_added'])})
        if not self.connection_3.strategy.sync:
            _, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertTrue(result['description'], 'success')
        self.assertTrue(b'sn_added' in self.connection_3.strategy.entries[dn]['sn'])

    def test_modify_add_nonexisting_singlevalue_1(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_1.bind()
        result = self.connection_1.modify(dn, {'title': (MODIFY_ADD, ['title_added'])})
        if not self.connection_1.strategy.sync:
            _, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertTrue(result['description'], 'success')
        self.assertTrue(b'title_added' in self.connection_1.strategy.entries[dn]['title'])

    def test_modify_add_nonexisting_singlevalue_2(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_2.bind()
        result = self.connection_2.modify(dn, {'title': (MODIFY_ADD, ['title_added'])})
        if not self.connection_2.strategy.sync:
            _, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertTrue(result['description'], 'success')
        self.assertTrue(b'title_added' in self.connection_2.strategy.entries[dn]['title'])

    def test_modify_add_nonexisting_singlevalue_3(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_3.bind()
        result = self.connection_3.modify(dn, {'title': (MODIFY_ADD, ['title_added'])})
        if not self.connection_3.strategy.sync:
            _, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertTrue(result['description'], 'success')
        self.assertTrue(b'title_added' in self.connection_3.strategy.entries[dn]['title'])

    def test_modify_add_existing_multivalue_1(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_1.bind()
        result = self.connection_1.modify(dn, {'sn': (MODIFY_ADD, ['sn_added1', 'sn_added2'])})
        if not self.connection_1.strategy.sync:
            _, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertTrue(result['description'], 'success')
        self.assertTrue(b'sn_added1' in self.connection_1.strategy.entries[dn]['sn'])
        self.assertTrue(b'sn_added2' in self.connection_1.strategy.entries[dn]['sn'])

    def test_modify_add_existing_multivalue_2(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_2.bind()
        result = self.connection_2.modify(dn, {'sn': (MODIFY_ADD, ['sn_added1', 'sn_added2'])})
        if not self.connection_2.strategy.sync:
            _, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertTrue(result['description'], 'success')
        self.assertTrue(b'sn_added1' in self.connection_2.strategy.entries[dn]['sn'])
        self.assertTrue(b'sn_added2' in self.connection_2.strategy.entries[dn]['sn'])

    def test_modify_add_existing_multivalue_3(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_3.bind()
        result = self.connection_3.modify(dn, {'sn': (MODIFY_ADD, ['sn_added1', 'sn_added2'])})
        if not self.connection_3.strategy.sync:
            _, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertTrue(result['description'], 'success')
        self.assertTrue(b'sn_added1' in self.connection_3.strategy.entries[dn]['sn'])
        self.assertTrue(b'sn_added2' in self.connection_3.strategy.entries[dn]['sn'])

    def test_modify_add_nonexisting_multivalue_1(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_1.bind()
        result = self.connection_1.modify(dn, {'title': (MODIFY_ADD, ['title_added1', 'title_added2'])})
        if not self.connection_1.strategy.sync:
            _, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertTrue(result['description'], 'success')
        self.assertTrue(b'title_added1' in self.connection_1.strategy.entries[dn]['title'])
        self.assertTrue(b'title_added2' in self.connection_1.strategy.entries[dn]['title'])

    def test_modify_add_nonexisting_multivalue_2(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_2.bind()
        result = self.connection_2.modify(dn, {'title': (MODIFY_ADD, ['title_added1', 'title_added2'])})
        if not self.connection_2.strategy.sync:
            _, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertTrue(result['description'], 'success')
        self.assertTrue(b'title_added1' in self.connection_2.strategy.entries[dn]['title'])
        self.assertTrue(b'title_added2' in self.connection_2.strategy.entries[dn]['title'])

    def test_modify_add_nonexisting_multivalue_3(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_3.bind()
        result = self.connection_3.modify(dn, {'title': (MODIFY_ADD, ['title_added1', 'title_added2'])})
        if not self.connection_3.strategy.sync:
            _, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertTrue(result['description'], 'success')
        self.assertTrue(b'title_added1' in self.connection_3.strategy.entries[dn]['title'])
        self.assertTrue(b'title_added2' in self.connection_3.strategy.entries[dn]['title'])

    def test_modify_delete_nonexisting_attribute_1(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_1.bind()
        result = self.connection_1.modify(dn, {'initials': (MODIFY_DELETE, ['initials1', 'initials2'])})
        if not self.connection_1.strategy.sync:
            _, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertEqual(result['description'], 'noSuchAttribute')

    def test_modify_delete_nonexisting_attribute_2(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_2.bind()
        result = self.connection_2.modify(dn, {'initials': (MODIFY_DELETE, ['initials1', 'initials2'])})
        if not self.connection_2.strategy.sync:
            _, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertEqual(result['description'], 'noSuchAttribute')

    def test_modify_delete_nonexisting_attribute_3(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_3.bind()
        result = self.connection_3.modify(dn, {'initials': (MODIFY_DELETE, ['initials1', 'initials2'])})
        if not self.connection_3.strategy.sync:
            _, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertEqual(result['description'], 'noSuchAttribute')

    def test_modify_delete_existing_singlevalue_1(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_1.bind()
        result = self.connection_1.modify(dn, {'title': (MODIFY_DELETE, ['title1'])})
        if not self.connection_1.strategy.sync:
            _, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertEqual(result['description'], 'success')

    def test_modify_delete_existing_singlevalue_2(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_2.bind()
        result = self.connection_2.modify(dn, {'sn': (MODIFY_DELETE, ['user4_sn'])})
        if not self.connection_2.strategy.sync:
            _, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertEqual(result['description'], 'success')

    def test_modify_delete_existing_singlevalue_3(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_3.bind()
        result = self.connection_3.modify(dn, {'sn': (MODIFY_DELETE, ['user4_sn'])})
        if not self.connection_3.strategy.sync:
            _, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertEqual(result['description'], 'success')

    def test_modify_delete_existing_multivalue_1(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_1.bind()
        result = self.connection_1.modify(dn, {'title': (MODIFY_DELETE, ['title1', 'title2'])})
        if not self.connection_1.strategy.sync:
            _, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(self.connection_1.strategy.entries[dn]['title'] == [b'title3'])

    def test_modify_delete_existing_multivalue_2(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_2.bind()
        result = self.connection_2.modify(dn, {'title': (MODIFY_DELETE, ['title1', 'title2'])})
        if not self.connection_2.strategy.sync:
            _, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(self.connection_2.strategy.entries[dn]['title'] == [b'title3'])

    def test_modify_delete_existing_multivalue_3(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_3.bind()
        result = self.connection_3.modify(dn, {'title': (MODIFY_DELETE, ['title1', 'title2'])})
        if not self.connection_3.strategy.sync:
            _, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(self.connection_3.strategy.entries[dn]['title'] == [b'title3'])

    def test_modify_replace_existing_singlevalue_1(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_1.bind()
        result = self.connection_1.modify(dn, {'sn': (MODIFY_REPLACE, ['user_test_sn'])})
        if not self.connection_1.strategy.sync:
            _, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(b'user_test_sn' in self.connection_1.strategy.entries[dn]['sn'])
        self.assertEqual(len(self.connection_1.strategy.entries[dn]['sn']), 1)

    def test_modify_replace_existing_singlevalue_2(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_2.bind()
        result = self.connection_2.modify(dn, {'sn': (MODIFY_REPLACE, ['user_test_sn'])})
        if not self.connection_2.strategy.sync:
            _, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(b'user_test_sn' in self.connection_2.strategy.entries[dn]['sn'])
        self.assertEqual(len(self.connection_2.strategy.entries[dn]['sn']), 1)

    def test_modify_replace_existing_singlevalue_3(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_3.bind()
        result = self.connection_3.modify(dn, {'sn': (MODIFY_REPLACE, ['user_test_sn'])})
        if not self.connection_3.strategy.sync:
            _, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(b'user_test_sn' in self.connection_3.strategy.entries[dn]['sn'])
        self.assertEqual(len(self.connection_3.strategy.entries[dn]['sn']), 1)

    def test_modify_replace_existing_multivalue_1(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_1.bind()
        result = self.connection_1.modify(dn, {'title': (MODIFY_REPLACE, ['title4', 'title5'])})
        if not self.connection_1.strategy.sync:
            _, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertEqual(result['description'], 'success')
        self.assertFalse(b'title1' in self.connection_1.strategy.entries[dn]['title'])
        self.assertFalse(b'title2' in self.connection_1.strategy.entries[dn]['title'])
        self.assertFalse(b'title3' in self.connection_1.strategy.entries[dn]['title'])
        self.assertTrue(b'title4' in self.connection_1.strategy.entries[dn]['title'])
        self.assertTrue(b'title5' in self.connection_1.strategy.entries[dn]['title'])
        self.assertEqual(len(self.connection_1.strategy.entries[dn]['title']), 2)

    def test_modify_replace_existing_multivalue_2(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_2.bind()
        result = self.connection_2.modify(dn, {'title': (MODIFY_REPLACE, ['title4', 'title5'])})
        if not self.connection_2.strategy.sync:
            _, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertEqual(result['description'], 'success')
        self.assertFalse(b'title1' in self.connection_2.strategy.entries[dn]['title'])
        self.assertFalse(b'title2' in self.connection_2.strategy.entries[dn]['title'])
        self.assertFalse(b'title3' in self.connection_2.strategy.entries[dn]['title'])
        self.assertTrue(b'title4' in self.connection_2.strategy.entries[dn]['title'])
        self.assertTrue(b'title5' in self.connection_2.strategy.entries[dn]['title'])
        self.assertEqual(len(self.connection_2.strategy.entries[dn]['title']), 2)

    def test_modify_replace_existing_multivalue_3(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_3.bind()
        result = self.connection_3.modify(dn, {'title': (MODIFY_REPLACE, ['title4', 'title5'])})
        if not self.connection_3.strategy.sync:
            _, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertEqual(result['description'], 'success')
        self.assertFalse(b'title1' in self.connection_3.strategy.entries[dn]['title'])
        self.assertFalse(b'title2' in self.connection_3.strategy.entries[dn]['title'])
        self.assertFalse(b'title3' in self.connection_3.strategy.entries[dn]['title'])
        self.assertTrue(b'title4' in self.connection_3.strategy.entries[dn]['title'])
        self.assertTrue(b'title5' in self.connection_3.strategy.entries[dn]['title'])
        self.assertEqual(len(self.connection_3.strategy.entries[dn]['title']), 2)

    def test_modify_replace_existing_novalue_1(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_1.bind()
        result = self.connection_1.modify(dn, {'title': (MODIFY_REPLACE, [])})
        if not self.connection_1.strategy.sync:
            _, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertEqual(result['description'], 'success')
        self.assertFalse(b'title' in self.connection_1.strategy.entries[dn])

    def test_modify_replace_existing_novalue_2(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_2.bind()
        result = self.connection_2.modify(dn, {'title': (MODIFY_REPLACE, [])})
        if not self.connection_2.strategy.sync:
            _, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertEqual(result['description'], 'success')
        self.assertFalse(b'title' in self.connection_2.strategy.entries[dn])

    def test_modify_replace_existing_novalue_3(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_3.bind()
        result = self.connection_3.modify(dn, {'title': (MODIFY_REPLACE, [])})
        if not self.connection_3.strategy.sync:
            _, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertEqual(result['description'], 'success')
        self.assertFalse(b'title' in self.connection_3.strategy.entries[dn])

    def test_modify_replace_not_existing_novalue_1(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_1.bind()
        result = self.connection_1.modify(dn, {'initials': (MODIFY_REPLACE, [])})
        if not self.connection_1.strategy.sync:
            _, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertEqual(result['description'], 'success')
        self.assertFalse(b'initials' in self.connection_1.strategy.entries[dn])

    def test_modify_replace_not_existing_novalue_2(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_2.bind()
        result = self.connection_2.modify(dn, {'initials': (MODIFY_REPLACE, [])})
        if not self.connection_2.strategy.sync:
            _, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertEqual(result['description'], 'success')
        self.assertFalse(b'initials' in self.connection_2.strategy.entries[dn])

    def test_modify_replace_not_existing_novalue_3(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_3.bind()
        result = self.connection_3.modify(dn, {'initials': (MODIFY_REPLACE, [])})
        if not self.connection_3.strategy.sync:
            _, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertEqual(result['description'], 'success')
        self.assertFalse(b'initials' in self.connection_3.strategy.entries[dn])

    def test_modify_replace_not_existing_singlevalue_1(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_1.bind()
        result = self.connection_1.modify(dn, {'initials': (MODIFY_REPLACE, ['initials1'])})
        if not self.connection_1.strategy.sync:
            _, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(b'initials1' in self.connection_1.strategy.entries[dn]['initials'])

    def test_modify_replace_not_existing_singlevalue_2(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_2.bind()
        result = self.connection_2.modify(dn, {'initials': (MODIFY_REPLACE, ['initials1'])})
        if not self.connection_2.strategy.sync:
            _, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(b'initials1' in self.connection_2.strategy.entries[dn]['initials'])

    def test_modify_replace_not_existing_singlevalue_3(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_3.bind()
        result = self.connection_3.modify(dn, {'initials': (MODIFY_REPLACE, ['initials1'])})
        if not self.connection_3.strategy.sync:
            _, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(b'initials1' in self.connection_3.strategy.entries[dn]['initials'])

    def test_modify_replace_not_existing_multivalue_1(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_1.bind()
        result = self.connection_1.modify(dn, {'initials': (MODIFY_REPLACE, ['initials1', 'initials2'])})
        if not self.connection_1.strategy.sync:
            _, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(b'initials1' in self.connection_1.strategy.entries[dn]['initials'])
        self.assertTrue(b'initials2' in self.connection_1.strategy.entries[dn]['initials'])

    def test_modify_replace_not_existing_multivalue_2(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_2.bind()
        result = self.connection_2.modify(dn, {'initials': (MODIFY_REPLACE, ['initials1', 'initials2'])})
        if not self.connection_2.strategy.sync:
            _, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(b'initials1' in self.connection_2.strategy.entries[dn]['initials'])
        self.assertTrue(b'initials2' in self.connection_2.strategy.entries[dn]['initials'])

    def test_modify_replace_not_existing_multivalue_3(self):
        dn = 'cn=user4,ou=test,o=lab'
        self.connection_3.bind()
        result = self.connection_3.modify(dn, {'initials': (MODIFY_REPLACE, ['initials1', 'initials2'])})
        if not self.connection_3.strategy.sync:
            _, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(b'initials1' in self.connection_3.strategy.entries[dn]['initials'])
        self.assertTrue(b'initials2' in self.connection_3.strategy.entries[dn]['initials'])

    def test_search_exact_match_single_attribute_1(self):
        self.connection_1.bind()
        result = self.connection_1.search('o=lab', '(cn=user1)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_1.response
        if not self.connection_1.strategy.sync:
            response, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertEqual(result['description'], 'success')
        self.assertEqual('user1', response[0]['attributes']['cn'][0])

    def test_search_exact_match_single_attribute_2(self):
        self.connection_2.bind()
        result = self.connection_2.search('o=lab', '(cn=user2)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_2.response
        if not self.connection_2.strategy.sync:
            response, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertEqual(result['description'], 'success')
        self.assertEqual('user2', response[0]['attributes']['cn'][0])

    def test_search_exact_match_single_attribute_3(self):
        self.connection_3.bind()
        result = self.connection_3.search('o=lab', '(cn=user3)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_3.response
        if not self.connection_3.strategy.sync:
            response, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertEqual(result['description'], 'success')
        self.assertEqual('user3', response[0]['attributes']['cn'][0])

    def test_search_exact_match_case_insensitive_single_attribute_1(self):
        self.connection_1.bind()
        result = self.connection_1.search('o=lab', '(cn=UsEr1)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_1.response
        if not self.connection_1.strategy.sync:
            response, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertEqual(result['description'], 'success')
        self.assertEqual('user1', response[0]['attributes']['cn'][0])

    def test_search_exact_match_case_insensitive_single_attribute_2(self):
        self.connection_2.bind()
        result = self.connection_2.search('o=lab', '(cn=UsEr2)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_2.response
        if not self.connection_2.strategy.sync:
            response, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertEqual(result['description'], 'success')
        self.assertEqual('user2', response[0]['attributes']['cn'][0])

    def test_search_exact_match_case_insensitive_single_attribute_3(self):
        self.connection_3.bind()
        result = self.connection_3.search('o=lab', '(cn=UsEr3)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_3.response
        if not self.connection_3.strategy.sync:
            response, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertEqual(result['description'], 'success')
        self.assertEqual('user3', response[0]['attributes']['cn'][0])

    def test_search_presence_single_attribute_1(self):
        self.connection_1.bind()
        result = self.connection_1.search('o=lab', '(cn=*)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_1.response
        if not self.connection_1.strategy.sync:
            response, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue('user' in response[0]['attributes']['cn'][0])

    def test_search_presence_single_attribute_2(self):
        self.connection_2.bind()
        result = self.connection_2.search('o=lab', '(cn=*)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_2.response
        if not self.connection_2.strategy.sync:
            response, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue('user' in response[0]['attributes']['cn'][0])

    def test_search_presence_single_attribute_3(self):
        self.connection_3.bind()
        result = self.connection_3.search('o=lab', '(cn=*)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_3.response
        if not self.connection_3.strategy.sync:
            response, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue('user' in response[0]['attributes']['cn'][0])

    def test_search_presence_and_filter_1(self):
        self.connection_1.bind()
        result = self.connection_1.search('o=lab', '(&(cn=*)(|(sn=user0_sn)(sn=user1_sn)))', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_1.response
        if not self.connection_1.strategy.sync:
            response, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(response[0]['attributes']['cn'][0] in ['user0', 'user1'])

        self.connection_1b.bind()
        result = self.connection_1b.search('o=lab', '(&(cn=*)(|(sn=user0_sn)(sn=user1_sn)))', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_1b.response
        if not self.connection_1b.strategy.sync:
            response, result = self.connection_1b.get_response(result)
        else:
            result = self.connection_1b.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(response[0]['attributes']['cn'][0] in ['user0', 'user1'])

    def test_search_presence_and_filter_2(self):
        self.connection_2.bind()
        result = self.connection_2.search('o=lab', '(&(cn=*)(|(sn=user0_sn)(sn=user1_sn)))', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_2.response
        if not self.connection_2.strategy.sync:
            response, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(response[0]['attributes']['cn'][0] in ['user0', 'user1'])

        self.connection_2b.bind()
        result = self.connection_2b.search('o=lab', '(&(cn=*)(|(sn=user0_sn)(sn=user1_sn)))', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_2b.response
        if not self.connection_2b.strategy.sync:
            response, result = self.connection_2b.get_response(result)
        else:
            result = self.connection_2b.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(response[0]['attributes']['cn'][0] in ['user0', 'user1'])

    def test_search_presence_and_filter_3(self):
        self.connection_3.bind()
        result = self.connection_3.search('o=lab', '(&(cn=*)(|(sn=user0_sn)(sn=user1_sn)))', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_3.response
        if not self.connection_3.strategy.sync:
            response, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(response[0]['attributes']['cn'][0] in ['user0', 'user1'])

        self.connection_3b.bind()
        result = self.connection_3b.search('o=lab', '(&(cn=*)(|(sn=user0_sn)(sn=user1_sn)))', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_3b.response
        if not self.connection_3b.strategy.sync:
            response, result = self.connection_3b.get_response(result)
        else:
            result = self.connection_3b.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(response[0]['attributes']['cn'][0] in ['user0', 'user1'])

    def test_search_incorrect_base_1(self):
        self.connection_1.bind()
        result = self.connection_1.search('o=nonexistant', '(cn=*)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        if not self.connection_1.strategy.sync:
            _, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertEqual(result['description'], 'noSuchObject')

    def test_search_incorrect_base_2(self):
        self.connection_2.bind()
        result = self.connection_2.search('o=nonexistant', '(cn=*)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        if not self.connection_2.strategy.sync:
            _, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertEqual(result['description'], 'noSuchObject')

    def test_search_incorrect_base_3(self):
        self.connection_3.bind()
        result = self.connection_3.search('o=nonexistant', '(cn=*)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        if not self.connection_3.strategy.sync:
            _, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertEqual(result['description'], 'noSuchObject')

    def test_search_incorrect_base_exception_1(self):
        self.connection_1c.bind()
        try:
            result = self.connection_1c.search('o=nonexistant', '(cn=*)', search_scope=SUBTREE, attributes=['cn', 'sn'])
            self.fail('exception not raised')
        except LDAPNoSuchObjectResult:
            pass

        if not self.connection_1c.strategy.sync:
            _, result = self.connection_1c.get_response(result)
        else:
            result = self.connection_1c.result
        self.assertEqual(result['description'], 'noSuchObject')

    def test_search_incorrect_base_exception_2(self):
        self.connection_2c.bind()
        try:
            result = self.connection_2c.search('o=nonexistant', '(cn=*)', search_scope=SUBTREE, attributes=['cn', 'sn'])
            self.fail('exception not raised')
        except LDAPNoSuchObjectResult:
            pass

        if not self.connection_2c.strategy.sync:
            _, result = self.connection_2c.get_response(result)
        else:
            result = self.connection_2c.result
        self.assertEqual(result['description'], 'noSuchObject')

    def test_search_incorrect_base_exception_3(self):
        self.connection_3c.bind()
        try:
            result = self.connection_3c.search('o=nonexistant', '(cn=*)', search_scope=SUBTREE, attributes=['cn', 'sn'])
            self.fail('exception not raised')
        except LDAPNoSuchObjectResult:
            pass

        if not self.connection_3c.strategy.sync:
            _, result = self.connection_3c.get_response(result)
        else:
            result = self.connection_3c.result
        self.assertEqual(result['description'], 'noSuchObject')

    def test_search_presence_and_filter_no_entries_found_1(self):
        self.connection_1.bind()
        result = self.connection_1.search('o=lab', '(&(cn=*)(sn=user_nonexistant))', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_1.response
        if not self.connection_1.strategy.sync:
            response, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertEqual(result['description'], 'success')
        self.assertEqual(len(response), 0)

    def test_search_presence_and_filter_no_entries_found_2(self):
        self.connection_2.bind()
        result = self.connection_2.search('o=lab', '(&(cn=*)(sn=user_nonexistant))', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_2.response
        if not self.connection_2.strategy.sync:
            response, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertEqual(result['description'], 'success')
        self.assertEqual(len(response), 0)

    def test_search_presence_and_filter_no_entries_found_3(self):
        self.connection_3.bind()
        result = self.connection_3.search('o=lab', '(&(cn=*)(sn=user_nonexistant))', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_3.response
        if not self.connection_3.strategy.sync:
            response, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertEqual(result['description'], 'success')
        self.assertEqual(len(response), 0)

    def test_search_exact_match_not_filter_1(self):
        self.connection_1.bind()
        result = self.connection_1.search('o=lab', '(!(sn=user0_sn))', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_1.response
        if not self.connection_1.strategy.sync:
            response, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertEqual(result['description'], 'success')
        self.assertNotEqual(response[0]['attributes']['cn'][0], 'user0')

    def test_search_exact_match_not_filter_2(self):
        self.connection_2.bind()
        result = self.connection_2.search('o=lab', '(!(sn=user0_sn))', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_2.response
        if not self.connection_2.strategy.sync:
            response, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertEqual(result['description'], 'success')
        self.assertNotEqual(response[0]['attributes']['cn'][0], 'user0')

    def test_search_exact_match_not_filter_3(self):
        self.connection_3.bind()
        result = self.connection_3.search('o=lab', '(!(sn=user0_sn))', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_3.response
        if not self.connection_3.strategy.sync:
            response, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertEqual(result['description'], 'success')
        self.assertNotEqual(response[0]['attributes']['cn'][0], 'user0')

    def test_search_greater_or_equal_than_string_1(self):
        self.connection_1.bind()
        result = self.connection_1.search('o=lab', '(userPassword>=test2222)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_1.response
        if not self.connection_1.strategy.sync:
            response, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(response[0]['attributes']['cn'][0] in ['user2', 'user3', 'user4'])

    def test_search_greater_or_equal_than_string_2(self):
        self.connection_2.bind()
        result = self.connection_2.search('o=lab', '(userPassword>=test2222)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_2.response
        if not self.connection_2.strategy.sync:
            response, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(response[0]['attributes']['cn'][0] in ['user2', 'user3', 'user4'])

    def test_search_greater_or_equal_than_string_3(self):
        self.connection_3.bind()
        result = self.connection_3.search('o=lab', '(userPassword>=test2222)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_3.response
        if not self.connection_3.strategy.sync:
            response, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(response[0]['attributes']['cn'][0] in ['user2', 'user3', 'user4'])

    def test_search_greater_or_equal_than_int_1(self):
        self.connection_1.bind()
        result = self.connection_1.search('o=lab', '(revision>=2)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_1.response
        if not self.connection_1.strategy.sync:
            response, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(response[0]['attributes']['cn'][0] in ['user2', 'user3', 'user4'])

    def test_search_greater_or_equal_than_int_2(self):
        self.connection_2.bind()
        result = self.connection_2.search('o=lab', '(revision>=2)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_2.response
        if not self.connection_2.strategy.sync:
            response, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(response[0]['attributes']['cn'][0] in ['user2', 'user3', 'user4'])

    def test_search_greater_or_equal_than_int_3(self):
        self.connection_3.bind()
        result = self.connection_3.search('o=lab', '(revision>=2)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_3.response
        if not self.connection_3.strategy.sync:
            response, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(response[0]['attributes']['cn'][0] in ['user2', 'user3', 'user4'])

    def test_search_less_or_equal_than_string_1(self):
        self.connection_1.bind()
        result = self.connection_1.search('o=lab', '(userPassword<=test2222)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_1.response
        if not self.connection_1.strategy.sync:
            response, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(response[0]['attributes']['cn'][0] in ['user0', 'user1', 'user2'])

    def test_search_less_or_equal_than_string_2(self):
        self.connection_2.bind()
        result = self.connection_2.search('o=lab', '(userPassword<=test2222)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_2.response
        if not self.connection_2.strategy.sync:
            response, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(response[0]['attributes']['cn'][0] in ['user0', 'user1', 'user2'])

    def test_search_less_or_equal_than_string_3(self):
        self.connection_3.bind()
        result = self.connection_3.search('o=lab', '(userPassword<=test2222)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_3.response
        if not self.connection_3.strategy.sync:
            response, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(response[0]['attributes']['cn'][0] in ['user0', 'user1', 'user2'])

    def test_search_less_or_equal_than_int_1(self):
        self.connection_1.bind()
        result = self.connection_1.search('o=lab', '(revision<=2)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_1.response
        if not self.connection_1.strategy.sync:
            response, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(response[0]['attributes']['cn'][0] in ['user0', 'user1', 'user2'])

    def test_search_less_or_equal_than_int_2(self):
        self.connection_2.bind()
        result = self.connection_2.search('o=lab', '(revision<=2)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_2.response
        if not self.connection_2.strategy.sync:
            response, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(response[0]['attributes']['cn'][0] in ['user0', 'user1', 'user2'])

    def test_search_less_or_equal_than_int_3(self):
        self.connection_3.bind()
        result = self.connection_3.search('o=lab', '(revision<=2)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_3.response
        if not self.connection_3.strategy.sync:
            response, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(response[0]['attributes']['cn'][0] in ['user0', 'user1', 'user2'])

    def test_search_substring_1(self):
        self.connection_1.bind()
        result = self.connection_1.search('o=lab', '(cn=*ser*2)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_1.response
        if not self.connection_1.strategy.sync:
            response, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(response[0]['attributes']['cn'][0] in ['user2'])

    def test_search_substring_2(self):
        self.connection_2.bind()
        result = self.connection_2.search('o=lab', '(cn=*ser*2)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_2.response
        if not self.connection_2.strategy.sync:
            response, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(response[0]['attributes']['cn'][0] in ['user2'])

    def test_search_substring_3(self):
        self.connection_3.bind()
        result = self.connection_3.search('o=lab', '(cn=*ser*2)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_3.response
        if not self.connection_3.strategy.sync:
            response, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(response[0]['attributes']['cn'][0] in ['user2'])

    def test_search_case_insensitive_substring_1(self):
        self.connection_1.bind()
        result = self.connection_1.search('o=lab', '(cn=*SeR*2)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_1.response
        if not self.connection_1.strategy.sync:
            response, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(response[0]['attributes']['cn'][0] in ['user2'])

    def test_search_case_insensitive_substring_2(self):
        self.connection_2.bind()
        result = self.connection_2.search('o=lab', '(cn=*SeR*2)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_2.response
        if not self.connection_2.strategy.sync:
            response, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(response[0]['attributes']['cn'][0] in ['user2'])

    def test_search_case_insensitive_substring_3(self):
        self.connection_3.bind()
        result = self.connection_3.search('o=lab', '(cn=*SeR*2)', search_scope=SUBTREE, attributes=['cn', 'sn'])
        response = self.connection_3.response
        if not self.connection_3.strategy.sync:
            response, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(response[0]['attributes']['cn'][0] in ['user2'])

    def test_search_presence_singlevalue_attribute_1(self):
        self.connection_1.bind()
        result = self.connection_1.search('o=lab', '(revision=*)', search_scope=SUBTREE, attributes=['cn', 'revision'])
        response = self.connection_1.response
        if not self.connection_1.strategy.sync:
            response, result = self.connection_1.get_response(result)
        else:
            result = self.connection_1.result
        self.assertEqual(result['description'], 'success')

        self.assertTrue(isinstance(response[0]['attributes']['revision'], int))

    def test_search_presence_singlevalue_attribute_2(self):
        self.connection_2.bind()
        result = self.connection_2.search('o=lab', '(revision=*)', search_scope=SUBTREE, attributes=['cn', 'revision'])
        response = self.connection_2.response
        if not self.connection_2.strategy.sync:
            response, result = self.connection_2.get_response(result)
        else:
            result = self.connection_2.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(isinstance(response[0]['attributes']['revision'], int))

    def test_search_presence_singlevalue_attribute_3(self):
        self.connection_3.bind()
        result = self.connection_3.search('o=lab', '(revision=*)', search_scope=SUBTREE, attributes=['cn', 'revision'])
        response = self.connection_3.response
        if not self.connection_3.strategy.sync:
            response, result = self.connection_3.get_response(result)
        else:
            result = self.connection_3.result
        self.assertEqual(result['description'], 'success')
        self.assertTrue(isinstance(response[0]['attributes']['revision'], list))  # no schema so attributes are returned as lists

    def test_auth_ad(self):
        import ldap3
        mock_server = ldap3.Server('ldaps://server.domain.local:636', get_info=ldap3.OFFLINE_AD_2012_R2)
        mock_ldap_connection = ldap3.Connection(
            mock_server, client_strategy=ldap3.MOCK_SYNC, authentication=ldap3.SIMPLE)
        ldap_entries = 'ad_ldap_directory.json'
        mock_ldap_connection.strategy.entries_from_json(ldap_entries)
        mock_ldap_connection.open()
        mock_ldap_connection.user = 'CN=Test User,OU=PostMaster,DC=postmaster,DC=local'
        mock_ldap_connection.password = 'P@ssW0rd'
        result = mock_ldap_connection.bind()
        self.assertTrue(result)
        mock_ldap_connection.search('DC=postmaster,DC=local', '(distinguishedName=CN=Test User,OU=PostMaster,DC=postmaster,DC=local)', attributes=['sAMAccountName', 'objectSid'])
        username = mock_ldap_connection.response[0]['attributes']['sAMAccountName']
        self.assertEqual(len(mock_ldap_connection.response), 1)
        self.assertEqual(username, 'testUser')
        mock_ldap_connection.search('DC=postmaster,DC=local', '(objectSid=S-1-5-21-1270288957-3800934213-3019856503-1608)', attributes=['sAMAccountName'])
        username = mock_ldap_connection.response[0]['attributes']['sAMAccountName']
        self.assertEqual(len(mock_ldap_connection.response), 1)
        self.assertEqual(username, 'testUser')
        mock_ldap_connection.search('DC=postmaster,DC=local', '(sAMAccountName=testUser)', attributes=['sAMAccountName'])
        username = mock_ldap_connection.response[0]['attributes']['sAMAccountName']
        self.assertEqual(len(mock_ldap_connection.response), 1)
        self.assertEqual(username, 'testUser')
        mock_ldap_connection.search('DC=postmaster,DC=local', '(sAMAccountName=TeStUsER)', attributes=['sAMAccountName'])
        username = mock_ldap_connection.response[0]['attributes']['sAMAccountName']
        self.assertEqual(len(mock_ldap_connection.response), 1)
        self.assertEqual(username, 'testUser')


"""
"""

# Created on 2015.02.01
#
# Author: Giovanni Cannata
#
# Copyright 2015, 2016 Giovanni Cannata
#
# This file is part of ldap3.
#
# ldap3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ldap3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ldap3 in the COPYING and COPYING.LESSER files.
# If not, see <http://www.gnu.org/licenses/>.

import unittest

from ldap3 import SUBTREE
from ldap3.utils.conv import escape_bytes

from test import test_base, test_name_attr, random_id, get_connection, \
    add_user, drop_connection, test_server_type, test_int_attr


testcase_id = random_id()


class Test(unittest.TestCase):
    def setUp(self):
        self.connection = get_connection()
        self.delete_at_teardown = []
        if test_server_type == 'EDIR':
            self.delete_at_teardown.append(add_user(self.connection, testcase_id, 'search-1', attributes={'givenName': 'givenname-1', test_int_attr: 0}))
            self.delete_at_teardown.append(add_user(self.connection, testcase_id, 'search-2', attributes={'givenName': 'givenname-2', test_int_attr: 0}))
        elif test_server_type == 'AD':
            self.delete_at_teardown.append(add_user(self.connection, testcase_id, 'search-1', attributes={'givenName': 'givenname-1'}))
            self.delete_at_teardown.append(add_user(self.connection, testcase_id, 'search-2', attributes={'givenName': 'givenname-2'}))
        else:
            self.delete_at_teardown.append(add_user(self.connection, testcase_id, 'search-1', attributes={'givenName': 'givenname-1'}))
            self.delete_at_teardown.append(add_user(self.connection, testcase_id, 'search-2', attributes={'givenName': 'givenname-2'}))

    def tearDown(self):
        drop_connection(self.connection, self.delete_at_teardown)
        self.assertFalse(self.connection.bound)

    def test_search_exact_match(self):
        result = self.connection.search(search_base=test_base, search_filter='(' + test_name_attr + '=' + testcase_id + 'search-1)', attributes=[test_name_attr, 'givenName'])
        if not self.connection.strategy.sync:
            response, result = self.connection.get_response(result)
            entries = self.connection._get_entries(response)
        else:
            result = self.connection.result
            entries = self.connection.entries
        self.assertEqual(result['description'], 'success')
        self.assertEqual(len(entries), 1)
        self.assertEqual(entries[0].givenName.value, 'givenname-1')
        self.assertEqual(entries[0].givenname.value, 'givenname-1')
        self.assertEqual(entries[0].GIVENNAME.value, 'givenname-1')

    def test_search_extensible_match(self):
        if test_server_type == 'EDIR' and not self.connection.strategy.no_real_dsa:
            result = self.connection.search(search_base=test_base, search_filter='(&(ou:dn:=fixtures)(objectclass=inetOrgPerson))', attributes=[test_name_attr, 'givenName', 'sn'])
            if not self.connection.strategy.sync:
                response, result = self.connection.get_response(result)
                entries = self.connection._get_entries(response)
            else:
                result = self.connection.result
                entries = self.connection.entries
            self.assertEqual(result['description'], 'success')
            self.assertTrue(len(entries) >= 2)

    def test_search_present(self):
        result = self.connection.search(search_base=test_base, search_filter='(' + test_name_attr + '=*)', search_scope=SUBTREE, attributes=[test_name_attr, 'givenName'])
        if not self.connection.strategy.sync:
            response, result = self.connection.get_response(result)
            entries = self.connection._get_entries(response)
        else:
            result = self.connection.result
            entries = self.connection.entries
        self.assertEqual(result['description'], 'success')
        self.assertTrue(len(entries) >= 2)

    def test_search_substring_many(self):
        result = self.connection.search(search_base=test_base, search_filter='(' + test_name_attr + '=' + testcase_id + '*)', attributes=[test_name_attr, 'givenName'])
        if not self.connection.strategy.sync:
            response, result = self.connection.get_response(result)
            entries = self.connection._get_entries(response)
        else:
            result = self.connection.result
            entries = self.connection.entries
        self.assertEqual(result['description'], 'success')
        self.assertEqual(len(entries), 2)

    def test_search_with_operational_attributes(self):
        result = self.connection.search(search_base=test_base, search_filter='(' + test_name_attr + '=' + testcase_id + 'search-1)', search_scope=SUBTREE, attributes=[test_name_attr, 'givenName'], get_operational_attributes=True)
        if not self.connection.strategy.sync:
            response, result = self.connection.get_response(result)
            entries = self.connection._get_entries(response)
        else:
            result = self.connection.result
            entries = self.connection.entries
        self.assertEqual(result['description'], 'success')
        if self.connection.check_names:
            self.assertEqual(entries[0].entry_dn.lower(), self.delete_at_teardown[0][0].lower())

    def test_search_simple_paged(self):
        if not self.connection.strategy.pooled:
            self.delete_at_teardown.append(add_user(self.connection, testcase_id, 'search-3', attributes={'givenName': 'givenname-3'}))
            self.delete_at_teardown.append(add_user(self.connection, testcase_id, 'search-4', attributes={'givenName': 'givenname-4'}))
            self.delete_at_teardown.append(add_user(self.connection, testcase_id, 'search-5', attributes={'givenName': 'givenname-5'}))
            self.delete_at_teardown.append(add_user(self.connection, testcase_id, 'search-6', attributes={'givenName': 'givenname-6'}))
            self.delete_at_teardown.append(add_user(self.connection, testcase_id, 'search-7', attributes={'givenName': 'givenname-7'}))
            self.delete_at_teardown.append(add_user(self.connection, testcase_id, 'search-8', attributes={'givenName': 'givenname-8'}))
            self.delete_at_teardown.append(add_user(self.connection, testcase_id, 'search-9', attributes={'givenName': 'givenname-9'}))

            paged_size = 4
            total_entries = 0
            result = self.connection.search(search_base=test_base, search_filter='(' + test_name_attr + '=' + testcase_id + '*)', search_scope=SUBTREE, attributes=[test_name_attr, 'givenName'], paged_size=paged_size)
            if not self.connection.strategy.sync:
                response, result = self.connection.get_response(result)
                entries = self.connection._get_entries(response)
            else:
                result = self.connection.result
                entries = self.connection.entries
            self.assertEqual(result['description'], 'success')
            self.assertEqual(len(entries), paged_size)
            total_entries += len(entries)
            cookie = result['controls']['1.2.840.113556.1.4.319']['value']['cookie']
            while cookie:
                result = self.connection.search(search_base=test_base, search_filter='(' + test_name_attr + '=' + testcase_id + '*)', search_scope=SUBTREE, attributes=[test_name_attr, 'givenName'], paged_size=paged_size, paged_cookie=cookie)
                if not self.connection.strategy.sync:
                    response, result = self.connection.get_response(result)
                    entries = self.connection._get_entries(response)
                else:
                    result = self.connection.result
                    entries = self.connection.entries
                self.assertEqual(result['description'], 'success')
                total_entries += len(entries)
                self.assertTrue(len(entries) <= paged_size)
                cookie = result['controls']['1.2.840.113556.1.4.319']['value']['cookie']
            self.assertEqual(total_entries, 9)

    def test_search_exact_match_with_parentheses_in_filter(self):
        self.delete_at_teardown.append(add_user(self.connection, testcase_id, '(search)-10', attributes={'givenName': 'givenname-10'}))
        result = self.connection.search(search_base=test_base, search_filter='(' + test_name_attr + '=' + testcase_id + '*' + escape_bytes(')') + '*)', attributes=[test_name_attr, 'sn'])
        if not self.connection.strategy.sync:
            response, result = self.connection.get_response(result)
            entries = self.connection._get_entries(response)
        else:
            result = self.connection.result
            entries = self.connection.entries
        self.assertEqual(result['description'], 'success')
        self.assertEqual(len(entries), 1)
        self.assertEqual(entries[0][test_name_attr][0], testcase_id + '(search)-10')

    def test_search_integer_exact_match(self):
        result = self.connection.search(search_base=test_base, search_filter='(&(' + test_name_attr + '=' + testcase_id + '*)(' + test_int_attr + '=0))', attributes=[test_name_attr, test_int_attr])
        if not self.connection.strategy.sync:
            response, result = self.connection.get_response(result)
            entries = self.connection._get_entries(response)
        else:
            result = self.connection.result
            entries = self.connection.entries
        self.assertEqual(result['description'], 'success')
        self.assertEqual(len(entries), 2)

    def test_search_integer_less_than(self):
        result = self.connection.search(search_base=test_base, search_filter='(&(' + test_name_attr + '=' + testcase_id + '*)(' + test_int_attr + ' <=1))', attributes=[test_name_attr, test_int_attr])
        if not self.connection.strategy.sync:
            response, result = self.connection.get_response(result)
            entries = self.connection._get_entries(response)
        else:
            result = self.connection.result
            entries = self.connection.entries
        self.assertEqual(result['description'], 'success')
        self.assertEqual(len(entries), 2)

    def test_search_integer_greater_than(self):
        result = self.connection.search(search_base=test_base, search_filter='(&(' + test_name_attr + '=' + testcase_id + '*)(' + test_int_attr + ' >=-1))', attributes=[test_name_attr, test_int_attr])
        if not self.connection.strategy.sync:
            response, result = self.connection.get_response(result)
            entries = self.connection._get_entries(response)
        else:
            result = self.connection.result
            entries = self.connection.entries
        self.assertEqual(result['description'], 'success')
        self.assertEqual(len(entries), 2)



"""
"""

# Created on 2014.02.02
#
# Author: Giovanni Cannata
#
# Copyright 2014, 2015, 2016 Giovanni Cannata
#
# This file is part of ldap3.
#
# ldap3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ldap3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ldap3 in the COPYING and COPYING.LESSER files.
# If not, see <http://www.gnu.org/licenses/>.

import unittest

from test import get_connection, drop_connection


class Test(unittest.TestCase):
    def setUp(self):
        self.connection = get_connection(bind=False)

    def tearDown(self):
        drop_connection(self.connection)
        self.assertFalse(self.connection.bound)

    def test_open_connection(self):
        self.connection.open()
        self.assertEqual(self.connection.closed, False)
        self.connection.unbind()
        if self.connection.strategy.pooled:
            self.connection.strategy.terminate()
        self.assertEqual(self.connection.closed, True)
        self.assertEqual(self.connection.bound, False)

    def test_bind_connection(self):
        self.connection.open()
        self.assertEqual(self.connection.closed, False)
        self.connection.bind()
        self.assertEqual(self.connection.bound, True)
        self.connection.unbind()
        if self.connection.strategy.pooled:
            self.connection.strategy.terminate()
        self.assertEqual(self.connection.closed, True)
        self.assertEqual(self.connection.bound, False)

    def test_connection_in_context(self):
        with self.connection:
            self.assertEqual(self.connection.closed, False)
            self.assertEqual(self.connection.bound, True)

        if self.connection.strategy.pooled:
            self.connection.strategy.terminate()

        self.assertEqual(self.connection.closed, True)
        self.assertEqual(self.connection.bound, False)

    def test_connection_in_context_with_as(self):
        with self.connection as c:
            self.assertEqual(c.closed, False)
            self.assertEqual(c.bound, True)

        if self.connection.strategy.pooled:
            self.connection.strategy.terminate()

        self.assertEqual(self.connection.closed, True)
        self.assertEqual(self.connection.bound, False)



"""
"""

# Created on 2013.06.06
#
# Author: Giovanni Cannata
#
# Copyright 2013, 2014, 2015, 2016 Giovanni Cannata
#
# This file is part of ldap3.
#
# ldap3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ldap3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ldap3 in the COPYING and COPYING.LESSER files.
# If not, see <http://www.gnu.org/licenses/>.

import unittest

from ldap3 import ANONYMOUS, SASL, NTLM, DIGEST_MD5
from test import test_sasl_user, test_sasl_password, random_id, get_connection, drop_connection, test_sasl_realm, \
    test_server_type, test_ntlm_user, test_ntlm_password, test_secondary_user, test_secondary_password, \
    test_sasl_secondary_user, test_sasl_secondary_password, test_sasl_secondary_user_dn, test_sasl_user_dn

testcase_id = random_id()


class Test(unittest.TestCase):
    def test_bind_clear_text_to_secondary_user(self):
        connection = get_connection()
        self.assertTrue(connection.bound)
        connection.rebind(test_secondary_user, test_secondary_password)
        if test_server_type == 'EDIR':
            bound_dn = connection.extend.novell.get_bind_dn()
        else:
            bound_dn = connection.extend.standard.who_am_i()

        self.assertTrue(test_secondary_user in bound_dn)

        drop_connection(connection)
        self.assertFalse(connection.bound)

    def test_bind_anonymous_to_secondary_user(self):
        connection = get_connection(bind=True, lazy_connection=False, authentication=ANONYMOUS)
        self.assertTrue(connection.bound)
        connection.rebind(test_secondary_user, test_secondary_password)
        if test_server_type == 'EDIR':
            bound_dn = connection.extend.novell.get_bind_dn()
        else:
            bound_dn = connection.extend.standard.who_am_i()

        self.assertTrue(test_secondary_user in bound_dn)

        drop_connection(connection)
        self.assertFalse(connection.bound)

    def test_bind_sasl_digest_md5_to_secondary_sasl_user(self):
        connection = get_connection(bind=False, authentication=SASL, sasl_mechanism=DIGEST_MD5, sasl_credentials=(test_sasl_realm, test_sasl_user, test_sasl_password, None))
        connection.open()
        connection.bind()
        self.assertTrue(connection.bound)
        if test_server_type == 'EDIR':
            connected_user = connection.extend.novell.get_bind_dn()
        else:
            connected_user = str(connection.extend.standard.who_am_i())
        self.assertEqual(connected_user, test_sasl_user_dn)

        if connection.rebind(authentication=SASL, sasl_mechanism=DIGEST_MD5, sasl_credentials=(test_sasl_realm, test_sasl_secondary_user, test_sasl_secondary_password, None)):
            if test_server_type == 'EDIR':
                connected_user = connection.extend.novell.get_bind_dn()
            else:
                connected_user = connection.extend.standard.who_am_i()

            self.assertEqual(connected_user, test_sasl_secondary_user_dn)
        else:
            self.fail('secondary user sasl authentication failed')

        drop_connection(connection)
        self.assertFalse(connection.bound)

    def test_ntlm(self):
        if test_server_type == 'AD':
            connection = get_connection(bind=False, authentication=NTLM, ntlm_credentials=(test_ntlm_user, test_ntlm_password))
            connection.open()
            connection.bind()
            self.assertTrue(connection.bound)
            connected_user = str(connection.extend.standard.who_am_i())[2:]
            self.assertEqual(connected_user, test_ntlm_user)
            drop_connection(connection)
            self.assertFalse(connection.bound)



"""
"""

# Created on 2014.01.12
#
# Author: Giovanni Cannata
#
# Copyright 2014, 2015, 2016 Giovanni Cannata
#
# This file is part of ldap3.
#
# ldap3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ldap3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ldap3 in the COPYING and COPYING.LESSER files.
# If not, see <http://www.gnu.org/licenses/>.

import unittest

from ldap3.utils.ciDict import CaseInsensitiveDict


class Test(unittest.TestCase):
    def test_create_empty_case_insensitive_dict(self):
        cid = CaseInsensitiveDict()
        self.assertTrue(isinstance(cid, CaseInsensitiveDict))

    def test_create_case_insensitive_dict_from_dict(self):
        dic = dict()
        dic['ONE'] = 1
        dic['TWO'] = 2
        dic[3] = 3
        cid = CaseInsensitiveDict(dic)
        self.assertEqual(cid['ONE'], 1)
        self.assertEqual(cid['one'], 1)
        self.assertEqual(cid['TWO'], 2)
        self.assertEqual(cid['two'], 2)
        self.assertEqual(cid[3], 3)

    def test_create_case_insensitive_dict_from_parameters(self):
        cid = CaseInsensitiveDict(one=1, two=2)
        self.assertEqual(cid['ONE'], 1)
        self.assertEqual(cid['one'], 1)
        self.assertEqual(cid['TWO'], 2)
        self.assertEqual(cid['two'], 2)

    def test_add_values_to_case_insentitive_dict(self):
        cid = CaseInsensitiveDict()
        cid['oNe'] = 1
        cid['tWo'] = 2
        cid[3] = 3
        self.assertEqual(cid['ONE'], 1)
        self.assertEqual(cid['one'], 1)
        self.assertEqual(cid['TWO'], 2)
        self.assertEqual(cid['two'], 2)
        self.assertEqual(cid[3], 3)

    def test_modify_value_in_case_insentitive_dict_invariant_key(self):
        cid = CaseInsensitiveDict()
        cid['oNe'] = 1
        cid['tWo'] = 2
        cid[3] = 3
        self.assertEqual(cid[3], 3)
        cid[3] = 'Three'
        self.assertEqual(cid['ONE'], 1)
        self.assertEqual(cid['oNe'], 1)
        self.assertEqual(cid['TWO'], 2)
        self.assertEqual(cid['tWo'], 2)
        self.assertEqual(cid[3], 'Three')

    def test_modify_value_in_case_insentitive_dict_same_case_key(self):
        cid = CaseInsensitiveDict()
        cid['oNe'] = 1
        cid['tWo'] = 2
        cid[3] = 3
        self.assertEqual(cid['ONE'], 1)
        self.assertEqual(cid['oNe'], 1)
        cid['oNe'] = 'ONE'
        self.assertEqual(cid['ONE'], 'ONE')
        self.assertEqual(cid['oNe'], 'ONE')
        self.assertEqual(cid['TWO'], 2)
        self.assertEqual(cid['two'], 2)
        self.assertEqual(cid[3], 3)

    def test_modify_value_in_case_insentitive_dict_different_case_key(self):
        cid = CaseInsensitiveDict()
        cid['oNe'] = 1
        cid['tWo'] = 2
        cid[3] = 3
        self.assertEqual(cid['ONE'], 1)
        self.assertEqual(cid['oNe'], 1)
        cid['one'] = 'ONE'
        self.assertEqual(cid['ONE'], 'ONE')
        self.assertEqual(cid['oNe'], 'ONE')
        self.assertEqual(cid['TWO'], 2)
        self.assertEqual(cid['two'], 2)
        self.assertEqual(cid[3], 3)

    def test_delete_item_in_case_insentitive_dict_invariant_key(self):
        cid = CaseInsensitiveDict()
        cid['oNe'] = 1
        cid['tWo'] = 2
        cid[3] = 3
        self.assertEqual(cid[3], 3)
        del cid[3]
        self.assertEqual(cid['ONE'], 1)
        self.assertEqual(cid['oNe'], 1)
        self.assertEqual(cid['TWO'], 2)
        self.assertEqual(cid['tWo'], 2)
        try:
            cid[3]
        except KeyError:
            self.assertTrue(True)
        except Exception:
            self.assertTrue(False)

    def test_delete_item_in_case_insentitive_dict_same_case_key(self):
        cid = CaseInsensitiveDict()
        cid['oNe'] = 1
        cid['tWo'] = 2
        cid[3] = 3
        self.assertEqual(cid['ONE'], 1)
        self.assertEqual(cid['oNe'], 1)
        del cid['oNe']
        self.assertEqual(cid['TWO'], 2)
        self.assertEqual(cid['two'], 2)
        self.assertEqual(cid[3], 3)

        try:
            cid['oNe']
        except KeyError:
            self.assertTrue(True)
        except Exception:
            self.assertTrue(False)

        try:
            cid['ONE']
        except KeyError:
            self.assertTrue(True)
        except Exception:
            self.assertTrue(False)

    def test_delete_item_in_case_insentitive_dict_different_case_key(self):
        cid = CaseInsensitiveDict()
        cid['oNe'] = 1
        cid['tWo'] = 2
        cid[3] = 3
        self.assertEqual(cid['ONE'], 1)
        self.assertEqual(cid['oNe'], 1)
        del cid['one']
        self.assertEqual(cid['TWO'], 2)
        self.assertEqual(cid['two'], 2)
        self.assertEqual(cid[3], 3)
        try:
            cid['oNe']
        except KeyError:
            self.assertTrue(True)
        except Exception:
            self.assertTrue(False)

        try:
            cid['ONE']
        except KeyError:
            self.assertTrue(True)
        except Exception:
            self.assertTrue(False)

    def test_len_empty_case_insensitive_dict(self):
        cid = CaseInsensitiveDict()
        self.assertEqual(len(cid), 0)

    def test_len_case_insentitive_dict(self):
        cid = CaseInsensitiveDict()
        cid['oNe'] = 1
        cid['tWo'] = 2
        cid[3] = 3
        self.assertEqual(len(cid), 3)
        cid['ONE'] = 'ONE'
        self.assertEqual(len(cid), 3)

    def test_case_insensitive_dict_contains_invariant_key(self):
        cid = CaseInsensitiveDict()
        cid['oNe'] = 1
        cid['tWo'] = 2
        cid[3] = 3
        self.assertTrue(3 in cid)
        self.assertFalse('THREE' in cid)
        self.assertFalse(4 in cid)

    def test_case_insensitive_dict_contains_same_case_key(self):
        cid = CaseInsensitiveDict()
        cid['oNe'] = 1
        cid['tWo'] = 2
        cid[3] = 3
        self.assertTrue('oNe' in cid)
        self.assertFalse('THREE' in cid)
        self.assertFalse(4 in cid)

    def test_case_insensitive_dict_contains_different_case_key(self):
        cid = CaseInsensitiveDict()
        cid['oNe'] = 1
        cid['tWo'] = 2
        cid[3] = 3
        self.assertTrue('ONE' in cid)
        self.assertFalse('THREE' in cid)
        self.assertFalse(4 in cid)

    def test_copy_case_insensitive_dict(self):
        cid = CaseInsensitiveDict()
        cid['oNe'] = 1
        cid['tWo'] = 2
        cid[3] = 3

        cid2 = cid.copy()
        self.assertEqual(cid2['ONE'], 1)
        self.assertEqual(cid2['one'], 1)
        self.assertEqual(cid2['TWO'], 2)
        self.assertEqual(cid2['two'], 2)
        self.assertEqual(cid2[3], 3)

    def test_equality_case_insensitive_dict_with_same_case(self):
        cid = CaseInsensitiveDict()
        cid['one'] = 1
        cid['two'] = 2
        cid[3] = 3

        cid2 = CaseInsensitiveDict()
        cid2['one'] = 1
        cid2['two'] = 2
        cid2[3] = 3

        self.assertEqual(cid, cid2)

    def test_equality_case_insensitive_dict_with_different_case(self):
        cid = CaseInsensitiveDict()
        cid['one'] = 1
        cid['two'] = 2
        cid[3] = 3

        cid2 = CaseInsensitiveDict()
        cid2['ONE'] = 1
        cid2['TWO'] = 2
        cid2[3] = 3

        self.assertEqual(cid, cid2)

    def test_equality_case_insensitive_dict_with_same_case_dict(self):
        cid = CaseInsensitiveDict()
        cid['one'] = 1
        cid['two'] = 2
        cid[3] = 3

        dic = dict()
        dic['one'] = 1
        dic['two'] = 2
        dic[3] = 3

        self.assertEqual(cid, dic)

    def test_equality_case_insensitive_dict_with_different_case_dict(self):
        cid = CaseInsensitiveDict()
        cid['one'] = 1
        cid['two'] = 2
        cid[3] = 3

        dic = dict()
        dic['ONE'] = 1
        dic['TWO'] = 2
        dic[3] = 3

        self.assertEqual(cid, dic)

    def test_preserve_key_case_case_insensitive_dict(self):
        cid = CaseInsensitiveDict()
        cid['One'] = 1
        cid['Two'] = 2
        cid[3] = 3
        key_list = list(cid.keys())
        self.assertTrue('One' in key_list)
        self.assertTrue('Two' in key_list)
        self.assertTrue(3 in key_list)
        self.assertFalse('ONE' in key_list)
        self.assertFalse('one' in key_list)
        self.assertFalse('TWO' in key_list)
        self.assertFalse('TWO' in key_list)
        self.assertFalse(4 in key_list)



"""
"""

# Created on 2016.04.14
#
# Author: Giovanni Cannata
#
# Copyright 2016 Giovanni Cannata
#
# This file is part of ldap3.
#
# ldap3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ldap3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ldap3 in the COPYING and COPYING.LESSER files.
# If not, see <http://www.gnu.org/licenses/>.

from ...extend.operation import ExtendedOperation
from ...protocol.novell import EndGroupTypeRequestValue, EndGroupTypeResponseValue, Sequence
from ...utils.asn1 import decoder


class EndTransaction(ExtendedOperation):
    def config(self):
        self.request_name = '2.16.840.1.113719.1.27.103.2'
        self.response_name = '2.16.840.1.113719.1.27.103.2'
        self.request_value = EndGroupTypeRequestValue()
        self.asn1_spec = EndGroupTypeResponseValue()

    def __init__(self, connection, commit=True, controls=None):
        if controls and len(controls) == 1:
            group_cookie = decoder.decode(controls[0][2], asn1Spec=Sequence())[0][0]  # get the cookie from the built groupingControl
        else:
            group_cookie = None
        controls = None

        ExtendedOperation.__init__(self, connection, controls)  # calls super __init__()
        if group_cookie:
            self.request_value['endGroupCookie'] = group_cookie  # transactionGroupingType
            if not commit:
                self.request_value['endGroupValue'] = ''  # an empty endGroupValue means abort transaction

    def populate_result(self):
        try:
            self.result['value'] = self.decoded_response['endGroupValue']
        except TypeError:
            self.result['value'] = None

    def set_response(self):
        self.response_value = self.result



"""
"""

# Created on 2013.06.06
#
# Author: Giovanni Cannata
#
# Copyright 2013, 2014, 2015, 2016 Giovanni Cannata
#
# This file is part of ldap3.
#
# ldap3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ldap3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ldap3 in the COPYING and COPYING.LESSER files.
# If not, see <http://www.gnu.org/licenses/>.

import unittest
from time import sleep

from ldap3 import SUBTREE, MODIFY_ADD, MODIFY_REPLACE, MODIFY_DELETE, SIMPLE
from ldap3.protocol.microsoft import extended_dn_control, show_deleted_control
from test import test_base, test_name_attr, random_id, get_connection, add_user, drop_connection, test_server_type, test_root_partition

testcase_id = random_id()


class Test(unittest.TestCase):
    def setUp(self):
        if test_server_type == 'AD':
            self.connection = get_connection(use_ssl=True)
            self.delete_at_teardown = []
            # creates 2000 users for testing auto-range
            # for i in range(2000):
            #     try:
            #         add_user(self.connection, '', 'user-' + str(i).zfill(4), attributes={'givenName': 'givenname-' + str(i).zfill(4)})
            #         print(i)
            #     except:
            #         pass
            # self.delete_at_teardown.append(add_user(self.connection, testcase_id, 'search-2', attributes={'givenName': 'givenname-2'}))

    def tearDown(self):
        if test_server_type == 'AD':
            # drop_connection(self.connection, self.delete_at_teardown)
            drop_connection(self.connection, False)
            self.assertFalse(self.connection.bound)

    def test_search_extended_dn_ad(self):
        if test_server_type == 'AD':
            result = self.connection.search(search_base=test_base, search_filter='(' + test_name_attr + '=' + testcase_id + 'search-1)', attributes=[test_name_attr], controls=[extended_dn_control(), show_deleted_control()])
            if not self.connection.strategy.sync:
                response, result = self.connection.get_response(result)
            else:
                response = self.connection.response
                result = self.connection.result

            self.assertEqual(result['description'], 'success')
            self.assertTrue('<GUID=' in response[0]['dn'])
            self.assertTrue('SID=' in response[0]['dn'])
            self.assertTrue('>;' in response[0]['dn'])

    def test_search_deleted_objects_ad(self):
        if test_server_type == 'AD':
            dn_to_delete, _ = add_user(self.connection, testcase_id, 'to-be-deleted-1', attributes={'givenName': 'to-be-deleted-1'})
            sleep(1)
            self.connection.delete(dn_to_delete)
            sleep(1)
            result = self.connection.search(search_base=test_root_partition,
                                            search_filter='(&(isDeleted=TRUE)(cn=*' + testcase_id + '*deleted-1*))',
                                            search_scope=SUBTREE,
                                            attributes=[],
                                            controls=[show_deleted_control(criticality=True)])
            if not self.connection.strategy.sync:
                response, result = self.connection.get_response(result)
            else:
                response = self.connection.response
                result = self.connection.result
            found = False
            for entry in response:
                if entry['type'] == 'searchResEntry' and testcase_id in entry['dn']:
                    found = True
                    break

            self.assertTrue(found)

    def test_dir_sync(self):
        if test_server_type == 'AD':
            sync = self.connection.extend.microsoft.dir_sync(test_root_partition, attributes=['*'])
            # read all previous changes
            while sync.more_results:
                print('PREV', len(sync.loop()))

            # add a new object and verify the sync
            dn, _ = add_user(self.connection, testcase_id, 'to-be-deleted-1', attributes={'givenName': 'to-be-deleted-1'})
            sleep(1)
            response = sync.loop()
            print('ADD OBJ', len(response), response[0]['attributes'])
            found = False
            for entry in response:
                if entry['type'] == 'searchResEntry' and testcase_id + 'to-be-deleted-1' in entry['dn']:
                    found = True
                    break
            self.assertTrue(found)

            # modify-add an attribute and verify the sync
            result = self.connection.modify(dn, {'businessCategory': (MODIFY_ADD, ['businessCategory-1-added'])})
            if not self.connection.strategy.sync:
                _, result = self.connection.get_response(result)
            else:
                result = self.connection.result
            self.assertEqual(result['description'], 'success')
            sleep(1)
            response = sync.loop()
            print('MOD-ADD ATTR', len(response), response[0]['attributes'])
            found = False
            for entry in response:
                if entry['type'] == 'searchResEntry' and testcase_id + 'to-be-deleted-1' in entry['dn']:
                    found = True
                    break
            self.assertTrue(found)

            # modify-replace an attribute and verify the sync
            result = self.connection.modify(dn, {'businessCategory': (MODIFY_REPLACE, ['businessCategory-1-replaced']), 'sn': (MODIFY_REPLACE, ['sn-replaced'])})
            if not self.connection.strategy.sync:
                _, result = self.connection.get_response(result)
            else:
                result = self.connection.result
            self.assertEqual(result['description'], 'success')
            sleep(1)
            response = sync.loop()
            print('MOD-REPLACE ATTR', len(response), response[0]['attributes'])
            found = False
            for entry in response:
                if entry['type'] == 'searchResEntry' and testcase_id + 'to-be-deleted-1' in entry['dn']:
                    found = True
                    break
            self.assertTrue(found)

            # modify-delete an attribute and verify the sync
            # result = self.connection.modify(dn, {'businessCategory': (MODIFY_ADD, ['businessCategory-2-added', 'businessCategory-3-added'])})
            # if not self.connection.strategy.sync:
            #     _, result = self.connection.get_response(result)
            # else:
            #     result = self.connection.result
            # self.assertEqual(result['description'], 'success')
            result = self.connection.modify(dn, {'businessCategory': (MODIFY_DELETE, ['businessCategory-1-replaced'])})
            if not self.connection.strategy.sync:
                _, result = self.connection.get_response(result)
            else:
                result = self.connection.result
            self.assertEqual(result['description'], 'success')
            sleep(1)
            response = sync.loop()
            print('MOD-DEL ATTR', len(response), response[0]['attributes'])
            found = False
            for entry in response:
                if entry['type'] == 'searchResEntry' and testcase_id + 'to-be-deleted-1' in entry['dn']:
                    found = True
                    break
            self.assertTrue(found)

            # delete object and verify the sync
            self.connection.delete(dn)
            sleep(1)
            response = sync.loop()
            print('DEL OBJ', len(response), response[0]['attributes'])
            found = False
            for entry in response:
                if entry['type'] == 'searchResEntry' and testcase_id + 'to-be-deleted-1' in entry['dn']:
                    found = True
                    break

            self.assertTrue(found)

    def test_modify_password_as_administrator(self):
        if test_server_type == 'AD':
            self.delete_at_teardown.append(add_user(self.connection, testcase_id, 'changed-password-1', attributes={'givenName': 'changed-password-1'}))
            dn = self.delete_at_teardown[-1][0]
            # test_connection = get_connection(bind=False, authentication=SIMPLE, simple_credentials=(dn, 'Rc1234abcd'))
            # test_connection.bind()
            # self.assertTrue(test_connection.bound)
            # connected_user = test_connection.extend.standard.who_am_i()
            # test_connection.unbind()
            # self.assertTrue('changed-password-1' in connected_user)

            new_password = 'Rc5678efgh'
            result = self.connection.extend.microsoft.modify_password(dn, new_password)
            self.assertEqual(result, True)
            # creates a second connection and tries to bind with the new password
            test_connection = get_connection(bind=False, authentication=SIMPLE, simple_credentials=(dn, new_password))
            test_connection.bind()
            connected_user = test_connection.extend.standard.who_am_i()
            test_connection.unbind()

            self.assertTrue('changed-password-1' in connected_user)

    def test_modify_password_as_normal_user(self):
        if test_server_type == 'AD':
            old_password = 'Ab1234cdef'
            new_password = 'Gh5678ijkl'
            self.delete_at_teardown.append(add_user(self.connection, testcase_id, 'changed-password-2', password=old_password, attributes={'givenName': 'changed-password-2'}))
            dn = self.delete_at_teardown[-1][0]
            # creates a second connection and tries to bind with the new password
            test_connection = get_connection(bind=False, authentication=SIMPLE, simple_credentials=(dn, old_password))
            test_connection.bind()
            self.assertTrue(test_connection.bound)
            connected_user = test_connection.extend.standard.who_am_i()
            test_connection.unbind()
            self.assertTrue('changed-password-2' in connected_user)

            # changee the password
            result = self.connection.extend.microsoft.modify_password(dn, new_password, old_password)
            self.assertEqual(result, True)

            # tries to bind with the new password
            test_connection.password =  new_password
            test_connection.bind()
            connected_user = test_connection.extend.standard.who_am_i()
            test_connection.unbind()

            self.assertTrue('changed-password-2' in connected_user)

    def test_modify_existing_password_as_administrator(self):
        if test_server_type == 'AD':
            # self.delete_at_teardown.append(add_user(self.connection, testcase_id, 'changed-password-1', attributes={'givenName': 'changed-password-1'}))
            # dn = self.delete_at_teardown[-1][0]
            dn = 'CN=Test Uno,CN=Users,DC=TESTAD,DC=LAB'
            new_password = 'Rc579efgh'
            result = self.connection.extend.microsoft.modify_password(dn, new_password)
            self.assertEqual(result, True)
            # creates a second connection and tries to bind with the new password
            test_connection = get_connection(bind=False, authentication=SIMPLE, simple_credentials=(dn, new_password))
            test_connection.bind()
            connected_user = test_connection.extend.standard.who_am_i()
            test_connection.unbind()
            self.assertTrue('testuno' in connected_user)

    def test_search_with_auto_range(self):
        if test_server_type == 'AD':
            print(self.connection.auto_range)
            result = self.connection.search(search_base=test_base, search_filter='(' + test_name_attr + '=testgrp)', attributes=[test_name_attr, 'member'])
            print(result)
            print(self.connection.response[0])
            print (len(self.connection.response[0]['attributes']['member']))



"""
"""

# Created on 2016.04.30
#
# Author: Giovanni Cannata
#
# Copyright 2016 Giovanni Cannata
#
# This file is part of ldap3.
#
# ldap3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ldap3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ldap3 in the COPYING and COPYING.LESSER files.
# If not, see <http://www.gnu.org/licenses/>.

import json
import re

from threading import Lock

from .. import SEQUENCE_TYPES, ALL_ATTRIBUTES
from ..operation.bind import bind_request_to_dict
from ..operation.delete import delete_request_to_dict
from ..operation.add import add_request_to_dict
from ..operation.compare import compare_request_to_dict
from ..operation.modifyDn import modify_dn_request_to_dict
from ..operation.modify import modify_request_to_dict
from ..operation.search import search_request_to_dict, parse_filter, ROOT, AND, OR, NOT, MATCH_APPROX, \
    MATCH_GREATER_OR_EQUAL, MATCH_LESS_OR_EQUAL, MATCH_EXTENSIBLE, MATCH_PRESENT,\
    MATCH_SUBSTRING, MATCH_EQUAL
from ..utils.conv import json_hook, check_escape, to_unicode, to_raw
from ..core.exceptions import LDAPDefinitionError, LDAPPasswordIsMandatoryError
from ..utils.ciDict import CaseInsensitiveDict
from ..utils.dn import to_dn, safe_dn, safe_rdn
from ..protocol.sasl.sasl import validate_simple_password
from ..utils.log import log, log_enabled, ERROR, BASIC
from ..protocol.formatters.standard import format_attribute_values


# noinspection PyProtectedMember,PyUnresolvedReferences
class MockBaseStrategy(object):
    """
    Base class for connection strategy
    """

    def __init__(self):
        if not hasattr(self.connection.server, 'dit'):  # create entries dict if not already present
            self.connection.server.dit_lock = Lock()
            self.connection.server.dit = dict()
        self.entries = self.connection.server.dit  # for simpler reference
        self.no_real_dsa = True
        self.bound = None
        self.add_entry('cn=schema', [])  # add default entry for schema
        if log_enabled(BASIC):
            log(BASIC, 'instantiated <%s>: <%s>', self.__class__.__name__, self)

    def _start_listen(self):
        self.connection.listening = True
        self.connection.closed = False
        if self.connection.usage:
            self.connection._usage.open_sockets += 1

    def _stop_listen(self):
        self.connection.listening = False
        self.connection.closed = True
        if self.connection.usage:
            self.connection._usage.closed_sockets += 1

    def add_entry(self, dn, attributes):
        with self.connection.server.dit_lock:
            escaped_dn = safe_dn(dn)
            if escaped_dn not in self.connection.server.dit:
                self.connection.server.dit[escaped_dn] = CaseInsensitiveDict()
                for attribute in attributes:
                    if not isinstance(attributes[attribute], SEQUENCE_TYPES):  # entry attributes are always lists of bytes values
                        attributes[attribute] = [attributes[attribute]]
                    self.connection.server.dit[escaped_dn][attribute] = [to_raw(value) for value in attributes[attribute]]
                for rdn in safe_rdn(escaped_dn, decompose=True):  # adds rdns to entry attributes
                    if rdn[0] not in self.connection.server.dit[escaped_dn]:  # if rdn attribute is missing adds attribute and its value
                        self.connection.server.dit[escaped_dn][rdn[0]] = [to_raw(check_escape(rdn[1]))]
                    else:
                        if rdn[1] not in self.connection.server.dit[escaped_dn][rdn[0]]:  # add rdn value if rdn attribute is present but value is missing
                            self.connection.server.dit[escaped_dn][rdn[0]].append(to_raw(check_escape(rdn[1])))
                return True
            return False

    def remove_entry(self, dn):
        with self.connection.server.dit_lock:
            escaped_dn = safe_dn(dn)
            if escaped_dn in self.connection.server.dit:
                del self.connection.server.dit[escaped_dn]
                return True
            return False

    def entries_from_json(self, json_entry_file):
        target = open(json_entry_file, 'r')
        definition = json.load(target, object_hook=json_hook)
        if 'entries' not in definition:
            self.connection.last_error = 'invalid JSON definition, missing "entries" section'
            if log_enabled(ERROR):
                log(ERROR, '<%s> for <%s>', self.connection.last_error, self.connection)
            raise LDAPDefinitionError(self.connection.last_error)
        if not self.connection.server.dit:
            self.connection.server.dit = CaseInsensitiveDict()
        for entry in definition['entries']:
            if 'raw' not in entry:
                self.connection.last_error = 'invalid JSON definition, missing "raw" section'
                if log_enabled(ERROR):
                    log(ERROR, '<%s> for <%s>', self.connection.last_error, self.connection)
                raise LDAPDefinitionError(self.connection.last_error)
            if 'dn' not in entry:
                self.connection.last_error = 'invalid JSON definition, missing "dn" section'
                if log_enabled(ERROR):
                    log(ERROR, '<%s> for <%s>', self.connection.last_error, self.connection)
                raise LDAPDefinitionError(self.connection.last_error)
            self.add_entry(entry['dn'], entry['raw'])
        target.close()

    def mock_bind(self, request_message, controls):
        # BindRequest ::= [APPLICATION 0] SEQUENCE {
        #     version                 INTEGER (1 ..  127),
        #     name                    LDAPDN,
        #     authentication          AuthenticationChoice }
        #
        # BindResponse ::= [APPLICATION 1] SEQUENCE {
        #     COMPONENTS OF LDAPResult,
        #     serverSaslCreds    [7] OCTET STRING OPTIONAL }
        #
        # request: version, name, authentication
        # response: LDAPResult + serverSaslCreds
        request = bind_request_to_dict(request_message)
        identity = request['name']
        if 'simple' in request['authentication']:
            try:
                password = validate_simple_password(request['authentication']['simple'])
            except LDAPPasswordIsMandatoryError:
                password = ''
                identity = '<anonymous>'
        else:
            self.connection.last_error = 'only Simple Bind allowed in Mock strategy'
            if log_enabled(ERROR):
                log(ERROR, '<%s> for <%s>', self.connection.last_error, self.connection)
            raise LDAPDefinitionError(self.connection.last_error)
        # checks userPassword for password. userPassword must be a text string or a list of text strings
        if identity in self.connection.server.dit:
            if 'userPassword' in self.connection.server.dit[identity]:
                # if self.connection.server.dit[identity]['userPassword'] == password or password in self.connection.server.dit[identity]['userPassword']:
                if self.equal(identity, 'userPassword', password):
                    result_code = 0
                    message = ''
                    self.bound = identity
                else:
                    result_code = 49
                    message = 'invalid credentials'
            else:  # no user found, returns invalidCredentials
                result_code = 49
                message = 'missing userPassword attribute'
        elif identity == '<anonymous>':
            result_code = 0
            message = ''
            self.bound = identity
        else:
            result_code = 49
            message = 'missing object'

        return {'resultCode': result_code,
                'matchedDN': '',
                'diagnosticMessage': to_unicode(message, 'utf-8'),
                'referral': None,
                'serverSaslCreds': None
                }

    def mock_delete(self, request_message, controls):
        # DelRequest ::= [APPLICATION 10] LDAPDN
        #
        # DelResponse ::= [APPLICATION 11] LDAPResult
        #
        # request: entry
        # response: LDAPResult
        request = delete_request_to_dict(request_message)
        dn = safe_dn(request['entry'])
        if dn in self.connection.server.dit:
            del self.connection.server.dit[dn]
            result_code = 0
            message = ''
        else:
            result_code = 32
            message = 'object not found'

        return {'resultCode': result_code,
                'matchedDN': '',
                'diagnosticMessage': to_unicode(message, 'utf-8'),
                'referral': None
                }

    def mock_add(self, request_message, controls):
        # AddRequest ::= [APPLICATION 8] SEQUENCE {
        #     entry           LDAPDN,
        #     attributes      AttributeList }
        #
        # AddResponse ::= [APPLICATION 9] LDAPResult
        #
        # request: entry, attributes
        # response: LDAPResult
        request = add_request_to_dict(request_message)
        dn = safe_dn(request['entry'])
        attributes = request['attributes']
        # converts attributes values to bytes

        if dn not in self.connection.server.dit:
            if self.add_entry(dn, attributes):
                result_code = 0
                message = ''
            else:
                result_code = 1
                message = 'error adding entry'
        else:
            result_code = 68
            message = 'entry already exist'

        return {'resultCode': result_code,
                'matchedDN': '',
                'diagnosticMessage': to_unicode(message, 'utf-8'),
                'referral': None
                }

    def mock_compare(self, request_message, controls):
        # CompareRequest ::= [APPLICATION 14] SEQUENCE {
        #     entry           LDAPDN,
        #     ava             AttributeValueAssertion }
        #
        # CompareResponse ::= [APPLICATION 15] LDAPResult
        #
        # request: entry, attribute, value
        # response: LDAPResult
        request = compare_request_to_dict(request_message)
        dn = safe_dn(request['entry'])
        attribute = request['attribute']
        value = to_raw(request['value'])
        if dn in self.connection.server.dit:
            if attribute in self.connection.server.dit[dn]:
                if self.equal(dn, attribute, value):
                    result_code = 6
                    message = ''
                else:
                    result_code = 5
                    message = ''
            else:
                result_code = 16
                message = 'attribute not found'
        else:
            result_code = 32
            message = 'object not found'

        return {'resultCode': result_code,
                'matchedDN': '',
                'diagnosticMessage': to_unicode(message, 'utf-8'),
                'referral': None
                }

    def mock_modify_dn(self, request_message, controls):
        # ModifyDNRequest ::= [APPLICATION 12] SEQUENCE {
        #     entry           LDAPDN,
        #     newrdn          RelativeLDAPDN,
        #     deleteoldrdn    BOOLEAN,
        #     newSuperior     [0] LDAPDN OPTIONAL }
        #
        # ModifyDNResponse ::= [APPLICATION 13] LDAPResult
        #
        # request: entry, newRdn, deleteOldRdn, newSuperior
        # response: LDAPResult
        request = modify_dn_request_to_dict(request_message)
        dn = safe_dn(request['entry'])
        new_rdn = request['newRdn']
        delete_old_rdn = request['deleteOldRdn']
        new_superior = safe_dn(request['newSuperior']) if request['newSuperior'] else ''
        dn_components = to_dn(dn)
        if dn in self.connection.server.dit:
            if new_superior and new_rdn:  # performs move in the DIT
                self.connection.server.dit[safe_dn(dn_components[0] + ',' + new_superior)] = self.connection.server.dit[dn].copy()
                if delete_old_rdn:
                    del self.connection.server.dit[dn]
                result_code = 0
                message = 'entry moved'
            elif new_rdn and not new_superior:  # performs rename
                self.connection.server.dit[safe_dn(new_rdn + ',' + safe_dn(dn_components[1:]))] = self.connection.server.dit[dn].copy()
                del self.connection.server.dit[dn]
                result_code = 0
                message = 'entry rdn renamed'
            else:
                result_code = 53
                message = 'newRdn or newSuperior missing'
        else:
            result_code = 32
            message = 'object not found'

        return {'resultCode': result_code,
                'matchedDN': '',
                'diagnosticMessage': to_unicode(message, 'utf-8'),
                'referral': None
                }

    def mock_modify(self, request_message, controls):
        # ModifyRequest ::= [APPLICATION 6] SEQUENCE {
        #     object          LDAPDN,
        #     changes         SEQUENCE OF change SEQUENCE {
        #         operation       ENUMERATED {
        #             add     (0),
        #             delete  (1),
        #             replace (2),
        #             ...  },
        #         modification    PartialAttribute } }
        #
        # ModifyResponse ::= [APPLICATION 7] LDAPResult
        #
        # request: entry, changes
        # response: LDAPResult
        #
        # changes is a dictionary in the form {'attribute': [(operation, [val1, ...]), ...], ...}
        # operation is 0 (add), 1 (delete), 2 (replace), 3 (increment)
        request = modify_request_to_dict(request_message)
        dn = safe_dn(request['entry'])
        changes = request['changes']
        result_code = 0
        message = ''
        rdns = [rdn[0] for rdn in safe_rdn(dn, decompose=True)]
        if dn in self.connection.server.dit:
            original_entry = self.connection.server.dit[dn].copy()  # to preserve atomicity of operation
            for modification in changes:
                operation = modification['operation']
                attribute = modification['attribute']['type']
                elements = modification['attribute']['value']
                if operation == 0:  # add
                    if attribute not in self.connection.server.dit[dn] and elements:  # attribute not present, creates the new attribute and add elements
                        self.connection.server.dit[dn][attribute] = [to_raw(element) for element in elements]
                    else:  # attribute present, adds elements to current values
                        self.connection.server.dit[dn][attribute].extend([to_raw(element) for element in elements])
                elif operation == 1:  # delete
                    if attribute not in self.connection.server.dit[dn]:  # attribute must exist
                        result_code = 16
                        message = 'attribute must exists for deleting its values'
                    elif attribute in rdns:  # attribute can't be used in dn
                        result_code = 67
                        message = 'cannot delete an rdn'
                    else:
                        if not elements:  # deletes whole attribute if element list is empty
                            del self.connection.server.dit[dn][attribute]
                        else:
                            for element in elements:
                                raw_element = to_raw(element)
                                if self.equal(dn, attribute, raw_element):  # removes single element
                                    self.connection.server.dit[dn][attribute].remove(raw_element)
                                else:
                                    result_code = 1
                                    message = 'value to delete not found'
                            if not self.connection.server.dit[dn][attribute]:  # removes the whole attribute if no elements remained
                                del self.connection.server.dit[dn][attribute]
                elif operation == 2:  # replace
                    if attribute not in self.connection.server.dit[dn] and elements:  # attribute not present, creates the new attribute and add elements
                        self.connection.server.dit[dn][attribute] = [to_raw(element) for element in elements]
                    elif not elements and attribute in rdns:  # attribute can't be used in dn
                        result_code = 67
                        message = 'cannot replace an rdn'
                    elif not elements:  # deletes whole attribute if element list is empty
                        if attribute in self.connection.server.dit[dn]:
                            del self.connection.server.dit[dn][attribute]
                    else:  # substitutes elements
                        self.connection.server.dit[dn][attribute] = [to_raw(element) for element in elements]

            if result_code:  # an error has happened, restores the original dn
                self.connection.server.dit[dn] = original_entry
        else:
            result_code = 32
            message = 'object not found'

        return {'resultCode': result_code,
                'matchedDN': '',
                'diagnosticMessage': to_unicode(message, 'utf-8'),
                'referral': None
                }

    def mock_search(self, request_message, controls):
        # SearchRequest ::= [APPLICATION 3] SEQUENCE {
        #     baseObject      LDAPDN,
        #     scope           ENUMERATED {
        #         baseObject              (0),
        #         singleLevel             (1),
        #         wholeSubtree            (2),
        #     ...  },
        #     derefAliases    ENUMERATED {
        #         neverDerefAliases       (0),
        #         derefInSearching        (1),
        #         derefFindingBaseObj     (2),
        #         derefAlways             (3) },
        #     sizeLimit       INTEGER (0 ..  maxInt),
        #     timeLimit       INTEGER (0 ..  maxInt),
        #     typesOnly       BOOLEAN,
        #     filter          Filter,
        #     attributes      AttributeSelection }
        #
        # SearchResultEntry ::= [APPLICATION 4] SEQUENCE {
        #     objectName      LDAPDN,
        #     attributes      PartialAttributeList }
        #
        #
        # SearchResultReference ::= [APPLICATION 19] SEQUENCE
        #     SIZE (1..MAX) OF uri URI
        #
        # SearchResultDone ::= [APPLICATION 5] LDAPResult
        #
        # request: base, scope, dereferenceAlias, sizeLimit, timeLimit, typesOnly, filter, attributes
        # response_entry: object, attributes
        # response_done: LDAPResult
        request = search_request_to_dict(request_message)
        responses = []
        base = safe_dn(request['base'])
        scope = request['scope']
        attributes = request['attributes']
        filter_root = parse_filter(request['filter'], self.connection.server.schema, auto_escape=False)
        candidates = []
        if scope == 0:  # base object
            if base in self.connection.server.dit or base.lower() == 'cn=schema':
                candidates.append(base)
        elif scope == 1:  # single level
            for entry in self.connection.server.dit:
                if entry.endswith(base) and ',' not in entry[:-len(base) - 1]:  # only leafs without commas in the remaining dn
                    candidates.append(entry)
        elif scope == 2:  # whole subtree
            for entry in self.connection.server.dit:
                if entry.endswith(base):
                    candidates.append(entry)

        if not candidates:  # incorrect base
            result_code = 32
            message = 'incorrect base object'
        else:
            matched = self.evaluate_filter_node(filter_root, candidates)
            for match in matched:
                responses.append({
                    'object': match,
                    'attributes': [{'type': attribute,
                                    'vals': [] if request['typesOnly'] else self.connection.server.dit[match][attribute]}
                                   for attribute in self.connection.server.dit[match]
                                   if attribute in attributes or ALL_ATTRIBUTES in attributes]
                })

            result_code = 0
            message = ''

        result = {'resultCode': result_code,
                  'matchedDN': '',
                  'diagnosticMessage': to_unicode(message, 'utf-8'),
                  'referral': None
                  }

        return responses[:request['sizeLimit']] if request['sizeLimit'] > 0 else responses, result

    def evaluate_filter_node(self, node, candidates):
        """After evaluation each 2 sets are added to each MATCH node, one for the matched object and one for unmatched object.
        The unmatched object set is needed if a superior node is a NOT that reverts the evaluation. The BOOLEAN nodes mix the sets
        returned by the MATCH nodes"""
        node.matched = set()
        node.unmatched = set()

        if node.elements:
            for element in node.elements:
                self.evaluate_filter_node(element, candidates)

        if node.tag == ROOT:
            return node.elements[0].matched
        elif node.tag == AND:
            for element in node.elements:
                if not node.matched:
                    node.matched.update(element.matched)
                else:
                    node.matched.intersection_update(element.matched)
                if not node.unmatched:
                    node.unmatched.update(element.unmatched)
                else:
                    node.unmatched.intersection_update(element.unmatched)
        elif node.tag == OR:
            for element in node.elements:
                node.matched.update(element.matched)
                node.unmatched.update(element.unmatched)
        elif node.tag == NOT:
            node.matched = node.elements[0].unmatched
            node.unmatched = node.elements[0].matched
        elif node.tag == MATCH_GREATER_OR_EQUAL:
            attr_name = node.assertion['attr']
            attr_value = node.assertion['value']
            for candidate in candidates:
                if attr_name in self.connection.server.dit[candidate]:
                    for value in self.connection.server.dit[candidate][attr_name]:
                        if value.isdigit() and attr_value.isdigit():  # int comparison
                            if int(value) >= int(attr_value):
                                node.matched.add(candidate)
                            else:
                                node.unmatched.add(candidate)
                        else:
                            if to_unicode(value, 'utf-8').lower() >= to_unicode(attr_value, 'utf-8').lower():  # case insensitive string comparison
                                node.matched.add(candidate)
                            else:
                                node.unmatched.add(candidate)
        elif node.tag == MATCH_LESS_OR_EQUAL:
            attr_name = node.assertion['attr']
            attr_value = node.assertion['value']
            for candidate in candidates:
                if attr_name in self.connection.server.dit[candidate]:
                    for value in self.connection.server.dit[candidate][attr_name]:
                        if value.isdigit() and attr_value.isdigit():  # int comparison
                            if int(value) <= int(attr_value):
                                node.matched.add(candidate)
                            else:
                                node.unmatched.add(candidate)
                        else:
                            if to_unicode(value, 'utf-8').lower() <= to_unicode(attr_value, 'utf-8').lower():  # case insentive string comparison
                                node.matched.add(candidate)
                            else:
                                node.unmatched.add(candidate)
        elif node.tag == MATCH_EXTENSIBLE:
            self.connection.last_error = 'Extensible match not allowed in Mock strategy'
            if log_enabled(ERROR):
                log(ERROR, '<%s> for <%s>', self.connection.last_error, self.connection)
            raise LDAPDefinitionError(self.connection.last_error)
        elif node.tag == MATCH_PRESENT:
            attr_name = node.assertion['attr']
            for candidate in candidates:
                if attr_name in self.connection.server.dit[candidate]:
                    node.matched.add(candidate)
                else:
                    node.unmatched.add(candidate)
        elif node.tag == MATCH_SUBSTRING:
            attr_name = node.assertion['attr']
            # rebuild the original substring filter
            if node.assertion['initial']:
                substring_filter = re.escape(to_unicode(node.assertion['initial'], 'utf-8'))
            else:
                substring_filter = ''

            if node.assertion['any']:
                for middle in node.assertion['any']:
                    substring_filter += '.*' + re.escape(to_unicode(middle, 'utf-8'))

            if node.assertion['final']:
                substring_filter += '.*' + re.escape(to_unicode(node.assertion['final'], 'utf-8'))

            if substring_filter and not node.assertion['any'] and not node.assertion['final']:  # only initial, adds .*
                substring_filter += '.*'

            regex_filter = re.compile(substring_filter, flags=re.UNICODE | re.IGNORECASE)  # unicode AND ignorecase
            for candidate in candidates:
                if attr_name in self.connection.server.dit[candidate]:
                    for value in self.connection.server.dit[candidate][attr_name]:
                        if regex_filter.match(to_unicode(value, 'utf-8')):
                            node.matched.add(candidate)
                        else:
                            node.unmatched.add(candidate)
                else:
                    node.unmatched.add(candidate)
        elif node.tag == MATCH_EQUAL or node.tag == MATCH_APPROX:
            attr_name = node.assertion['attr']
            attr_value = node.assertion['value']
            for candidate in candidates:
                # if attr_name in self.connection.server.dit[candidate] and attr_value in self.connection.server.dit[candidate][attr_name]:
                if attr_name in self.connection.server.dit[candidate] and self.equal(candidate, attr_name, attr_value):
                    node.matched.add(candidate)
                # elif attr_name in self.connection.server.dit[candidate]:  # tries to apply formatters
                #     formatted_values = format_attribute_values(self.connection.server.schema, attr_name, self.connection.server.dit[candidate][attr_name], None)
                #     if not isinstance(formatted_values, SEQUENCE_TYPES):
                #         formatted_values = [formatted_values]
                #     # if attr_value.decode('utf-8') in formatted_values:  # attributes values should be returned in utf-8
                #     if self.equal(attr_name, attr_value.decode('utf-8'), formatted_values):  # attributes values should be returned in utf-8
                #         node.matched.add(candidate)
                #     else:
                #         node.unmatched.add(candidate)
                else:
                    node.unmatched.add(candidate)

    def equal(self, dn, attribute, value):
        # value is the value to match
        attribute_values = self.connection.server.dit[dn][attribute]
        if not isinstance(attribute_values, SEQUENCE_TYPES):
            attribute_values = [attribute_values]
        for attribute_value in attribute_values:
            if self._check_equality(value, attribute_value):
                return True

        # if not found tries to apply formatters
        formatted_values = format_attribute_values(self.connection.server.schema, attribute, attribute_values, None)
        if not isinstance(formatted_values, SEQUENCE_TYPES):
            formatted_values = [formatted_values]
        for attribute_value in formatted_values:
            if self._check_equality(value, attribute_value):
                return True

        return False

    @staticmethod
    def _check_equality(value1, value2):
        if value1.isdigit() and value2.isdigit():
            if int(value1) == int(value2):  # int comparison
                return True
        try:
            if to_unicode(value1, 'utf-8').lower() == to_unicode(value2, 'utf-8').lower():  # case insensitive comparison
                return True
        except UnicodeDecodeError:
            pass

        return False


"""
"""

# Created on 2016.08.31
#
# Author: Giovanni Cannata
#
# Copyright 2014, 2015, 2016 Giovanni Cannata
#
# This file is part of ldap3.
#
# ldap3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ldap3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with ldap3 in the COPYING and COPYING.LESSER files.
# If not, see <http://www.gnu.org/licenses/>.

STATUS_INIT = 'Initialized'  # The entry object is initialized
STATUS_VIRTUAL = 'Virtual'  # The entry is a new writable entry, still empty
STATUS_MANDATORY_MISSING = 'Missing mandatory attributes'  # The entry has some mandatory attributes missing
STATUS_READ = 'Read'  # The entry has been read
STATUS_WRITABLE = 'Writable'  # The entry has been made writable, still no changes
STATUS_PENDING_CHANGES = 'Pending changes'  # The entry has some changes to commit, mandatory attributes are present
STATUS_COMMITTED = 'Committed'  # The entry changes has been committed
STATUS_READY_FOR_DELETION = 'Ready for deletion'  # The entry is set to be deleted
STATUS_READY_FOR_MOVING = 'Ready for moving'  # The entry is set to be moved in the DIT
STATUS_READY_FOR_RENAMING = 'Ready for renaming'  # The entry is set to be renamed
STATUS_DELETED = 'Deleted'  # The entry has been deleted

STATUSES = [STATUS_INIT,
            STATUS_VIRTUAL,
            STATUS_MANDATORY_MISSING,
            STATUS_READ,
            STATUS_WRITABLE,
            STATUS_PENDING_CHANGES,
            STATUS_COMMITTED,
            STATUS_READY_FOR_DELETION,
            STATUS_READY_FOR_MOVING,
            STATUS_READY_FOR_RENAMING,
            STATUS_DELETED]

INITIAL_STATUSES = [STATUS_READ, STATUS_WRITABLE, STATUS_VIRTUAL]



# THIS FILE IS AUTO-GENERATED. PLEASE DO NOT MODIFY# version file for ldap3
# generated on 2016-11-18 08:08:03.830784
# on system uname_result(system='Windows', node='ELITE10GC', release='10', version='10.0.14393', machine='AMD64', processor='Intel64 Family 6 Model 58 Stepping 9, GenuineIntel')
# with Python 3.5.2 - ('v3.5.2:4def2a2901a5', 'Jun 25 2016 22:18:55') - MSC v.1900 64 bit (AMD64)
#
__version__ = '2.1.1'
__author__ = 'Giovanni Cannata'
__email__ = 'cannatag@gmail.com'
__url__ = 'https://github.com/cannatag/ldap3'
__description__ = 'A strictly RFC 4510 conforming LDAP V3 pure Python client library. The same codebase works with Python 2, Python 3, PyPy and PyPy3'
__status__ = '5 - Production/Stable'
__license__ = 'LGPL v3'
